#include <ctime>		// contains clock()
//#include <cmath>
using namespace std;

#include "vector.hpp"
#include "evpiter.cpp"
#include "flags.hpp"

#include "grid.hpp"
#include "grid2d.hpp"
#include "grid3d.hpp"
#include "laplace2d.hpp"
#include "laplace3d.hpp"
#include "multigrid.cpp"

/* --------------------------------------------------------------------- */

#define DEFAULT_LEVELS 4
#define DEFAULT_P      12
#define DEFAULT_ERRSTRAT 1
#define DEFAULT_A      1

/* --------------------------------------------------------------------- */
void InitVector2D(const unsigned int nx, double X[], double a);
void InitVector3D(const unsigned int nx, double X[], double a);


int main(const int argc, const char* argv[])
{
 unsigned int  levels, p, errstrat;
 int    i,nn, nx, ndim;
 double *delta, *c, *X;
 double tstart, tdiff, td;
 time_t t0,t1;
 double a;  // Konstante fuer Potentialfunktion

 if (argc <= 1)
   {
     cout << endl;
     cout << "-----------------------------------------------" <<endl;
     cout << "  Use the command line arguments in the following way" << endl << endl;
     cout << argv[0] << " [#levels] [#eigenfunctions] [strategy = 1/2] "
          << "[factor_in_helmholtz]" << endl << endl;
     cout << "  strategy=2 uses twice as much eigenfunctions internally" << endl;
     cout << "  factor = 0 results in the Laplacian" << endl << endl;
     cout << "-----------------------------------------------" <<endl;
     cout << endl;
   }

 levels = (argc > 1) ? atoi(argv[1]) : DEFAULT_LEVELS;
 if (levels<=0) exit (-1);
 p = (argc > 2) ? atoi(argv[2]) : DEFAULT_P;
 if (p<=0) exit (-2);
 errstrat = (argc > 3) ? atoi(argv[3]) : DEFAULT_ERRSTRAT;
 a        = (argc > 4) ? atof(argv[4]) : DEFAULT_A;

//  EvpIter<DiagLaplaceEqui3D> EVPsolv(levels, p);
// EvpIter<DiagLaplaceEqui3DNeumann> EVPsolv(levels, p);
// EvpIter<DiagLaplaceEqui3DRobin> EVPsolv(levels, p, true);
EvpIter<DiagLaplaceEqui2D> EVPsolv(levels, p, true);

 nn = EVPsolv.VectorLength();
 nx = EVPsolv.GetIntX();
 ndim = EVPsolv.NumberDimensions();

 cout << "nn = " << nn << ", nx = " << nx << ", ndim = " << ndim << endl;
 cout << " a = " << a << endl;

 // in 3D
 if (p>(nx-1)*(nx-1)*(nx-1))
  {
    cout << "Panic Stop: More eigenfunctions than Grid Points !!" << endl;
    exit(-1);
  }

 // anlegen der noetigen Vektoren
 delta = new double [2*p];
 c     = new double [nn];
 X     = new double [nn*p];

 // Vektor c = (0.0, ..., 0.0)
 //VDcopyC(nn, c, 0.0); 			// Helmholtz term !!

 switch(ndim)				// Initialize Eigenvectors
  {
    case 2: InitVectors2D(p,nx,X);
            InitVector2D(nx,c,a);
            break;
    case 3: InitVectors3D(p,nx,X);
            InitVector3D(nx,c,a);
            break;
    default: cout << "Dimension " << ndim << " is not valid." << endl;
             exit(-1);
  }

 //PrintVector(nn,c);


 // -----------------------------------------------------------------
 // --- Zeitmessung startet ...
 tstart = clock();
 t0 = time(0);


 // Aufruf der Iteration
 EVPsolv.Solve(c,delta,X,errstrat);


 t1 = time(0);
 td = difftime(t1,t0);
 tdiff  = (clock()-tstart)/CLOCKS_PER_SEC;


 // Ausgabe der Loesung
 for (i=0; i<p; i++)   delta[i] = 1.0/ delta[i];

#ifdef OUTPUT_EVP_Solution
 cout << "##############################################################" << endl << endl;
 cout << "  " << ndim << "D example with " << nn << " grid points ("
      << levels << " levels)" << endl;
 cout << "  " << p << " eigenfunctions calculated (" << errstrat*p
      << " are used internally)" << endl;

 PrintEValues(p, delta);
// PrintEFunctions3D(p,nx,X);
#endif


 cout << "Time spent in EVP-Solver for Helmholtz: " << tdiff << " sec. ( "
 	<< td << " sec.)" << endl;
 // --- ... Zeitmessung endet
 // -----------------------------------------------------------------

 // Speicher freigeben
 delete [] X;  delete [] delta; delete [] c;

 return 0;
}


int index2D(int i, int j, int nx)
{
   return ((nx+1)*i+j);
}

double f_hh2D(double x, double y, double a)
{
   double d = cosh(sqrt(a*(x*x+y*y)));
   return (1-1/(d*d));
}

// Helmholtzterm fuer harmonischen Oszillator
double f_hh_ho2D(const double x, const double y, const double a=1.0)
{
//  return 10;
   return (x*x+y*y)/2*a;
}


void InitVector2D(const unsigned int nx, double X[], double a)
{
   const double INV = 20;	// Intervall [-20,20]
   double x,y;
   int    i, j;

   for (i=0;i<=nx;i++)
   {
      x = 2*INV*( (i+0.0)/nx - 0.5 );
      for (j=0;j<=nx;j++)
      {
         y = 2*INV*( (j+0.0)/nx - 0.5 );
	 X[index2D(i,j,nx)] = f_hh_ho2D(x,y,a);
      }
   }
}


int index3D(int i, int j, int k, int nx)
{
   return ((nx+1)*(nx+1)*i+(nx+1)*j+k);
}


// Helmholtzterm fuer 1/cosh^2 - Potential
double f_hh3D(double x, double y, double z, double a)
{
   double d = cosh(a*sqrt(x*x+y*y+z*z));
   return (1-1/(d*d));
}


// Helmholtzterm fuer harmonischen Oszillator
double f_hh_ho3D(double x, double y, double z, double a)
{
   return (x*x+y*y+z*z)/2;
}


void InitVector3D(const unsigned int nx, double X[], double a)
{
  const double M2H  = 2 * 9.109534e-31 / 1.0545887e-34 / 1.0545887e-34;
  const double INTV = 20;
   int i, j, k;
   for (i=0;i<=nx;i++)
   {
      for (j=0;j<=nx;j++)
      {
         for (k=0;k<=nx;k++)
         {
// a
//            X[index3D(i,j,k,nx)] = 3*2*a*a*f_hh3D(INTV*(((double)k)/nx-0.5),INTV*(((double)j)/nx-0.5),INTV*(((double)i)/nx-0.5),a);
            X[index3D(i,j,k,nx)] = 3*f_hh3D(INTV*(((double)k)/nx-0.5),INTV*(((double)j)/nx-0.5),INTV*(((double)i)/nx-0.5),a);

// b
//            X[index3D(i,j,k,nx)] = f_hh_ho3D(100*(2*((double)k)/nx-1),100*(2*((double)j)/nx-1),100*(2*((double)i)/nx-1),a);
//            X[index3D(i,j,k,nx)] = f_hh_ho3D(INTV*((k+0.0)/nx-0.5),INTV*((j+0.0)/nx-0.5),INTV*((i+0.0)/nx-0.5),a);
//
            X[index3D(i,j,k,nx)] = 0.1;
         }
      }
   }
}

