#include <iostream>
//#include <cmath>
using namespace std;
#include "vector.hpp"
#include "grid.hpp"

//	###################################################################
//			Grid
//	###################################################################
//	general grid

Grid :: Grid(const int n)
 : length_(n), coarsegrid_(0), mat(0)
{}

//---------------------------------------------------------------------------
//	Attention:  pointer to coarse grid is copied but not the data

Grid :: Grid(const Grid &orig)
 : length_(orig.length_), coarsegrid_(orig.coarsegrid_), mat(0)
{
 if ( orig.mat != 0)
  {
    mat        = new double[length_*length_];
    VDcopy( length_*length_, mat, 1, orig.mat, 1 );
  }
//  else
//   {
//     mat = 0;
//   }
}

//---------------------------------------------------------------------------
Grid :: ~Grid()
{
 if (mat!=0)  delete [] mat;
}

//---------------------------------------------------------------------------

Grid& Grid :: operator=(const Grid & orig)
{
 if ( this != &orig )
 {
   if (mat != 0) { delete [] mat; };
   length_     = orig.length_;
   coarsegrid_ = orig.coarsegrid_;
   if ( orig.mat != 0)
    {
      mat = new double[length_*length_];
      VDcopy( length_*length_, mat, 1, orig.mat, 1 );
    }
   else
    {
      mat = 0;
    }
 }
 return *this;
}
//---------------------------------------------------------------------------

void Grid :: NewGridVector(double* &u) const
{
  u = new double[length_];
//   cout << " :: " << length_ << endl;
}

void Grid :: DeleteGridVector(double * &u) const
{
  if (u != 0)  { delete [] u;  u = 0; }
}

void Grid :: Inject(const double *uh, double *UH) const
{
 cout << endl << "Grid :: Inject is not implemented" << endl;
}

void Grid :: Restrict(const double *uh, double *UH) const
{
 cout << endl << "Grid :: Restrict is not implemented" << endl;
}

void Grid :: Interpolate(double *wh, const double *WH) const
{
 cout << endl << "Grid :: Interpolate is not implemented" << endl;
}

void Grid :: DefRes(const double *fh, const double *uh, double *DH) const
{
 cout << endl << "Grid :: DefRes is not implemented" << endl;
}

void Grid :: IntAdd(double *uh, const double *WH) const
{
 cout << endl << "Grid :: IntAdd is not implemented" << endl;
}

void Grid :: InvertMatrix()
{
 if ( mat != 0 )
  {
    cout << endl << "Grid :: InvertMatrix inverts matrix" << endl;
    MatrixInversion(VectorLength(), mat);
 }
 else
  {
    cout << endl << "Cannot invert an non-existent matrix in Grid :: InvertMatrix" << endl;
  }
 return;
}

void Grid :: DirectSolve(double *u, const double *f) const
{
  MatTimesVec(VectorLength(), mat, f, u);
}


void Grid :: DiagPC(double *w, const double *r) const
{
 cout << endl << "Grid :: DiagPC is not implemented" << endl;
}

void Grid :: Mult(double *f, const double *u) const
{
 cout << endl << "Grid :: Mult is not implemented" << endl;
}

void Grid :: Defect(const double *f, const double *u, double *def) const
{
 cout << endl << "Grid :: Defect is not implemented" << endl;
}

void Grid :: Smooth(double *u, const double *f, double *def,
  				const SmoothingType type, const int iter) const
{
 const double OMEGA=0.9;

#ifdef DEBUG
 double f1,f2;
 f1 = DefectNorm(u,f);
#endif

 switch(type)
   {
     case omega_Jacobi    : Jacobi(u,f,def,OMEGA,iter); break;
     case GS_forw   : GaussSeidelForward(u,f,iter);     break;
     case GS_backw  : GaussSeidelBackward(u,f,iter);    break;
     case ADI_forw  : ADIForward(u,f,def,iter);         break;
     case ADI_backw : ADIBackward(u,f,def,iter);        break;
   }
#ifdef DEBUG
 f2 = DefectNorm(u,f);
 cout << "n = " << VectorLength() << "  S("<< type <<") rate = " << f2/f1<< endl;
#endif
}

void Grid :: Jacobi(double *u, const double *f, double *def,
  			const double omega, const int iter) const
{
 cout << endl << "Grid ::  Jacobi is not implemented" << endl;
}

void Grid :: GaussSeidelForward(double *u, const double *f, const int iter) const
{
 cout << endl << "Grid :: GaussSeidelForward is not implemented" << endl;
}

void Grid :: GaussSeidelBackward(double *u, const double *f, const int iter) const
{
 cout << endl << "Grid :: GaussSeidelBackward is not implemented" << endl;
}

void Grid :: ADIForward(double *u, const double *f, double *aux, const int iter) const
{
 cout << endl << "Grid :: ADIForward is not implemented" << endl;
}

void Grid :: ADIBackward(double *u, const double *f, double *aux, const int iter) const
{
 cout << endl << "Grid :: ADIBacward is not implemented" << endl;
}

double Grid :: DefectNorm(const double *u, const double *f) const
{
 cout << endl << "Grid :: DefectNorm is not implemented" << endl;
 return 0.0;
}

void Grid :: SetBoundaryConst(double *u, const double c) const
{
 cout << endl << "Grid :: SetBoundaryConst is not implemented" << endl;
}

void Grid :: SetBoundaryZero(double *u) const
{
 cout << endl << "Grid :: SetBoundaryZero is not implemented" << endl;
}

//---------------------------------------------------------------------------:

