#include <iostream>
#include <cstdlib>
#include <fstream>		// needed for ifstream and ofstream
//#include <cmath>
using namespace std;
#include "vector.hpp"
#include "grid3d.hpp"

//	###################################################################
//			Grid3D
//	###################################################################
//	3D tensor product grid

Grid3D :: Grid3D(const int mx, const int my, const int mz)
 : Grid((mx+1)*(my+1)*(mz+1)),
   nx(mx), ny(my), nz(mz), dimflag(nx==ny && ny==nz),
   dim1(nx+1), dim2((nx+1)*(ny+1))
{
}

Grid3D :: Grid3D(const int mx)
 : Grid((mx+1)*(mx+1)*(mx+1)),
   nx(mx), ny(mx), nz(mx), dimflag(true),
   dim1(nx+1), dim2((nx+1)*(ny+1))
{
}

Grid3D :: ~Grid3D()
{}

//---------------------------------------------------------------------------

void Grid3D :: SetBoundaryZero(double u[]) const
{
 SetBoundaryConst(u, 0.0);
}

//---------------------------------------------------------------------------

void Grid3D :: SetBoundaryConst(double u[], const double c) const
{
 int ii,jj,n1,n2;

 n1 = nx+1;
 n2 = (nx+1)*(ny+1);

 for (ii=0; ii<=nx; ii++)
   for (jj=0; jj<=ny; jj++)
   {
      u[idx(n2,n1,  0, jj, ii )] = c;
      u[idx(n2,n1, nz, jj, ii )] = c;
   }
 for (ii=0; ii<=nx; ii++)
   for (jj=0; jj<=nz; jj++)
   {
      u[idx(n2,n1, jj,  0, ii )] = c;
      u[idx(n2,n1, jj, ny, ii )] = c;
   }
 for (ii=0; ii<=ny; ii++)
   for (jj=0; jj<=nz; jj++)
   {
      u[idx(n2,n1, jj, ii, 0 )]  = c;
      u[idx(n2,n1, jj, ii, nx )] = c;
   }
}
//---------------------------------------------------------------------------

void Grid3D :: AddBoundaryConst(double *u, const double c) const
{
 int ii,jj,n1,n2;

 n1 = nx+1;
 n2 = (nx+1)*(ny+1);

 for (ii=0; ii<=nx; ii++)
   for (jj=0; jj<=ny; jj++)
   {
      u[idx(n2,n1,  0, jj, ii )] += c;
      u[idx(n2,n1, nz, jj, ii )] += c;
   }
 for (ii=0; ii<=nx; ii++)
   for (jj=0; jj<=nz; jj++)
   {
      u[idx(n2,n1, jj,  0, ii )] += c;
      u[idx(n2,n1, jj, ny, ii )] += c;
   }
 for (ii=0; ii<=ny; ii++)
   for (jj=0; jj<=nz; jj++)
   {
      u[idx(n2,n1, jj, ii, 0 )]  += c;
      u[idx(n2,n1, jj, ii, nx )] += c;
   }
}
//---------------------------------------------------------------------------

void Grid3D :: CheckBoundaryZero(const double u[]) const
{
 int ii,jj,n1,n2;

 n1 = nx+1;
 n2 = (nx+1)*(ny+1);

 for (ii=0; ii<=nx; ii++)
   for (jj=0; jj<=ny; jj++)
   {
      if ( u[idx(n2,n1,  0, jj, ii )] != 0.0)
      	cout << " a Error" << endl;
      if ( u[idx(n2,n1, nz, jj, ii )] != 0.0)
      	cout << " a Error" << endl;
   }
 for (ii=0; ii<=nx; ii++)
   for (jj=0; jj<=nz; jj++)
   {
      if ( u[idx(n2,n1, jj,  0, ii )] != 0.0)
      	cout << " a Error" << endl;
      if ( u[idx(n2,n1, jj, ny, ii )] != 0.0)
      	cout << " a Error" << endl;
   }
 for (ii=0; ii<=ny; ii++)
   for (jj=0; jj<=nz; jj++)
   {
      if ( u[idx(n2,n1, jj, ii, 0 )]  != 0.0)
      	cout << " a Error" << endl;
      if ( u[idx(n2,n1, jj, ii, nx )] != 0.0)
      	cout << " a Error" << endl;
   }
}
//---------------------------------------------------------------------------

void Grid3D :: Inject(const double *uh, double *UH) const
{

  unsigned int	nf1,nf2,nc1,nc2,ncx,ncy,ncz,i,j,ii,jj,kk,ijk,IJK;
  const Grid3D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  ncz = coarsegrid.GetIntZ();
  nf1 = nx+1;   nf2 = (nx+1)*(ny+1);
  nc1 = ncx+1;  nc2 = (ncx+1)*(ncy+1);

  i = 2;
  for (ii=1; ii < ncz; ii++)
  {
    j = 2;
    for (jj=1; jj < ncy; jj++)
    {
      ijk = idx(nf2,nf1, i ,j , 2);
      IJK = idx(nc2,nc1, ii,jj, 1);
      for (kk=1; kk < ncx; kk++, IJK++)
      {
	UH[IJK] = uh[ijk];
	ijk += 2;
      }
      j += 2;
    }
    i += 2;
  }
  coarsegrid.SetBoundaryZero(UH);
}

//---------------------------------------------------------------------------:

void  Grid3D :: VectorToFile(const double u[], const char filename[]) const
{
				/* coordinates	  */
 double *xcoord = new double [VectorLength()*NumberDimensions()];
 GetCoords(xcoord);

 ofstream outfile(filename);
 if (outfile.fail())
 {
   cout << "File  " << filename << "  does not exist." << endl;
   exit(-1);
 }

 unsigned int ij=0;
 for (unsigned int k=0; k<=GetIntZ(); ++k)
 {
   for (unsigned int j=0; j<=GetIntY(); ++j)
   {
   for (unsigned int i=0; i<=GetIntX(); ++i, ++ij)
    {
     for (int s=0; s<NumberDimensions(); ++s)
     {
       outfile << xcoord[ij*NumberDimensions()+s] << "     ";
     }
     outfile << u[ij] << endl;
    }
    outfile << endl;
   }
 }

 delete [] xcoord;
 return;
}
//---------------------------------------------------------------------------
//	###################################################################
//			EquivGrid3D
//	###################################################################
//	Laplace 3D, equidistant tensor product grid

//---------------------------------------------------------------------------

Equi3D :: Equi3D(const int mx, const int my, const int mz)
 : Grid3D(mx,my,mz),
   hx(1.0/GetIntX()), hy(1.0/GetIntY()), hz(1.0/GetIntZ())
{
//   hx = 1.0/GetIntX();
//   hy = 1.0/GetIntY();
//   hz = 1.0/GetIntZ();
}

//---------------------------------------------------------------------------

Equi3D :: Equi3D(const int mx)
 : Grid3D(mx,mx,mx),
   hx(1.0/GetIntX()), hy(1.0/GetIntX()), hz(1.0/GetIntX())
{
//   hx = hy = hz = 1.0/GetIntX();
}

//---------------------------------------------------------------------------

Equi3D :: ~Equi3D()
{
}

//---------------------------------------------------------------------------

void Equi3D :: Refine(Equi3D & fine) const
{
  fine = Equi3D(2*nx,2*ny,2*nz);
  fine.coarsegrid_ = this;
}

//---------------------------------------------------------------------------

void Equi3D :: Restrict(const double uh[], double UH[]) const
{
//  const double  CW = 4./8.;			//  4 because h^2 -> rhs
  const double  CW = 1./8.;			//  normal weight

  double 	t,t2,t4,t8;
  unsigned int	nf1,nf2,nc1,nc2,ncx,ncy,ncz,i,j,k,ii,jj,kk,ijk,IJK;
  const Equi3D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  ncz = coarsegrid.GetIntZ();
  nf1 = nx+1;   nf2 = (nx+1)*(ny+1);
  nc1 = ncx+1;  nc2 = (ncx+1)*(ncy+1);

  VDcopyC(coarsegrid.VectorLength(), UH, 0.0e0);

  i = 2;
  for (ii=1; ii < ncz; ii++)
  {
    j = 2;
    for (jj=1; jj < ncy; jj++)
    {
      k = 2;
      ijk = idx(nf2,nf1, i ,j , k);
      IJK = idx(nc2,nc1, ii,jj, 1);
      for (kk=1; kk < ncx; kk++, IJK++)
      {
	t  = uh[ijk];

        t2 = ( uh[ijk-nf2] + uh[ijk+nf2] +
	       uh[ijk-nf1] + uh[ijk+nf1] +
	       uh[ijk-  1] + uh[ijk+  1]
	     )/2. ;

        t4 = ( uh[ijk-nf2-nf1] + uh[ijk-nf2+nf1] +
	       uh[ijk+nf2-nf1] + uh[ijk+nf2+nf1] +
	       uh[ijk-nf1-1]   + uh[ijk-nf1+1] +
	       uh[ijk+nf1-1]   + uh[ijk+nf1+1] +
	       uh[ijk-nf2-1]   + uh[ijk-nf2+1] +
	       uh[ijk+nf2-1]   + uh[ijk+nf2+1]
             )/4. ;

        t8 = ( uh[ijk-nf2-nf1-1] + uh[ijk-nf2-nf1+1] +
	       uh[ijk-nf2+nf1-1] + uh[ijk-nf2+nf1+1] +
               uh[ijk+nf2-nf1-1] + uh[ijk+nf2-nf1+1] +
	       uh[ijk+nf2+nf1-1] + uh[ijk+nf2+nf1+1]
	     )/8. ;

	UH[IJK] = (t+t2+t4+t8) * CW ;

	k += 2;  ijk += 2;
      }
      j += 2;
    }
    i += 2;
  }
//   CheckBoundaryZero(uh);
//   cout << "--------------------" << endl;
//   coarsegrid.CheckBoundaryZero(UH);
}

//---------------------------------------------------------------------------

void Equi3D :: Interpolate(double wh[], const double WH[]) const
{
 VDcopyC(VectorLength(), wh, 0.0);
 IntAdd(wh,WH);
}

//---------------------------------------------------------------------------

void Equi3D :: IntAdd(double uh[], const double WH[]) const
{
  double 	t,t2,t4,t8;
  unsigned int	nf1,nf2,nc1,nc2,ncx,ncy,ncz,i,j,k,ii,jj,kk,ijk,IJK;
  const Equi3D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  ncz = coarsegrid.GetIntZ();
  nf1 = nx+1;   nf2 = (nx+1)*(ny+1);
  nc1 = ncx+1;  nc2 = (ncx+1)*(ncy+1);

  i = 2;
  for (ii=1; ii < ncz; ii++)
  {
    j = 2;
    for (jj=1; jj < ncy; jj++)
    {
      k = 2;
      ijk = idx(nf2,nf1, i ,j , k);
      IJK = idx(nc2,nc1, ii,jj, 1);
      for (kk=1; kk < ncx; kk++, IJK++)
      {
	t  = WH[IJK];
	t2 = t/2.;
	t4 = t/4.;
	t8 = t/8.;

	uh[ijk]	+= t;

	uh[ijk-nf2]	+= t2;
	uh[ijk+nf2]	+= t2;
	uh[ijk-nf1]	+= t2;
	uh[ijk+nf1]	+= t2;
	uh[ijk-1]	+= t2;
	uh[ijk+1]	+= t2;

	uh[ijk-nf2-nf1]	+= t4;
	uh[ijk-nf2+nf1]	+= t4;
	uh[ijk+nf2-nf1]	+= t4;
	uh[ijk+nf2+nf1]	+= t4;
	uh[ijk-nf1-1]	+= t4;
	uh[ijk-nf1+1]	+= t4;
	uh[ijk+nf1-1]	+= t4;
	uh[ijk+nf1+1]	+= t4;
	uh[ijk-nf2-1]	+= t4;
	uh[ijk-nf2+1]	+= t4;
	uh[ijk+nf2-1]	+= t4;
	uh[ijk+nf2+1]	+= t4;

	uh[ijk-nf2-nf1-1]	+= t8;
	uh[ijk-nf2-nf1+1]	+= t8;
	uh[ijk-nf2+nf1-1]	+= t8;
	uh[ijk-nf2+nf1+1]	+= t8;
	uh[ijk+nf2-nf1-1]	+= t8;
	uh[ijk+nf2-nf1+1]	+= t8;
	uh[ijk+nf2+nf1-1]	+= t8;
	uh[ijk+nf2+nf1+1]	+= t8;

        k += 2;  ijk += 2;
      }
      j += 2;
    }
    i += 2;
  }
}
//---------------------------------------------------------------------------
/**	Calculates the coordinates of the given grid.
	The array @p xccord has to be allocated in advance
	(=VectorLength()*NumberDimensions())!!
	@param	xcoord	3d array of ccordinates.
*/
 void Equi3D :: GetCoords(double xcoord[]) const
 {
  unsigned int ijk = 0;
  for (unsigned int k=0; k<=GetIntZ(); ++k)
  {
    const double z0 = k*hz;
    for (unsigned int j=0; j<=GetIntY(); ++j)
    {
      const double y0 = j*hy;
      for (unsigned int i=0; i<=GetIntX(); ++i, ++ijk)
      {
        xcoord[3*ijk+0] = i*hx;
        xcoord[3*ijk+1] = y0;
        xcoord[3*ijk+2] = z0;
     }
    }
  }
  return;
 }
//---------------------------------------------------------------------------

//	###################################################################
//			NonEqui3D
//	###################################################################
/**	An non-equidistant 3D-tensor product grid
*/
/** 	Parameter constructor having numbers of nodes and
	node coordinates as parameters for all three directions.
*/
NonEqui3D :: NonEqui3D(
            const int nnx, const double x[],
            const int nny, const double y[],
	    const int nnz, const double z[])
 : Grid3D(nnx-1,nny-1,nnz-1),
   hhx(new double[GetIntX()+2]), hhy(new double[GetIntY()+2]),
   hhz(new double[GetIntZ()+2]),
   weight_xl(new double [GetIntX()+2]), weight_xr(new double [GetIntX()+2]),
   weight_yl(new double [GetIntY()+2]), weight_yr(new double [GetIntY()+2]),
   weight_zl(new double [GetIntZ()+2]), weight_zr(new double [GetIntZ()+2])

{
//   hhx = new double[GetIntX()+2]; // Index 1,..,GetIntX() is usefull
//   hhy = new double[GetIntY()+2];
//   hhz = new double[GetIntZ()+2];
				// Interval lengths
  				// Values at Indices 0 and GetIntX()+1 are set to zero
  hhx[0] = hhx[GetIntX()+1] = 0.0;
  for (int i=1; i<=GetIntX(); i++)
  	hhx[i] = x[i]-x[i-1];

  hhy[0] = hhy[GetIntY()+1] = 0.0;
  for (int i=1; i<=GetIntY(); i++)
  	hhy[i] = y[i]-y[i-1];

  hhz[0] = hhz[GetIntZ()+1] = 0.0;
  for (int i=1; i<=GetIntZ(); i++)
  	hhz[i] = z[i]-z[i-1];

//		Interpolation weights
//		We assume that all even nodes (0...nx+1) on the fine grid
//		are also nodes on the coarse grid

//   weight_xl = new double [GetIntX()+2];
//   weight_xr = new double [GetIntX()+2];
//   weight_yl = new double [GetIntY()+2];
//   weight_yr = new double [GetIntY()+2];
//   weight_zl = new double [GetIntZ()+2];
//   weight_zr = new double [GetIntZ()+2];

  CalcIntWeights();

  return;
}
//---------------------------------------------------------------------------

NonEqui3D :: NonEqui3D(const int nnx, const double *x)
 : Grid3D(nnx-1,nnx-1,nnx-1),hhx(0), hhy(0),hhz(0),
   weight_xl(0), weight_xr(0), weight_yl(0), weight_yr(0),
   weight_zl(0), weight_zr(0)
{
  if (nnx==1 || x==0)
   {
//      cout  << "No hhx-array allocated in NonEqui3D :: NonEqui3D " << endl;
   }
  else
   {
     hhx = new double[GetIntX()+2]; // Index 1,..,GetIntX() is usefull
     hhy = new double[GetIntY()+2];
     hhz = new double[GetIntZ()+2];
				// Interval lengths
  				// Values at Indices 0 and GetIntX()+1 are set to zero
     hhx[0] = hhx[GetIntX()+1] = 0.0;
     hhy[0] = hhy[GetIntY()+1] = 0.0;
     hhz[0] = hhz[GetIntZ()+1] = 0.0;
     for (int i=1; i<=GetIntX(); i++)
  	   hhx[i] = hhy[i] = hhz[i] = x[i]-x[i-1];

//		Interpolation weights
//		We assume that all even nodes (0...nx+1) on the fine grid
//		are also nodes on the coarse grid

     weight_xl = new double [GetIntX()+2];
     weight_xr = new double [GetIntX()+2];
     weight_yl = new double [GetIntY()+2];
     weight_yr = new double [GetIntY()+2];
     weight_zl = new double [GetIntZ()+2];
     weight_zr = new double [GetIntZ()+2];

     CalcIntWeights();
   }

  return;
}

//---------------------------------------------------------------------------
/** 	Parameter constructor having numbers of intervals and
	interval lengths as parameters for all three directions.
	xh[i] := x[i] - x[i-1] and
	xh[0:intx-1] is mapped onto hhx[1:intx]
	with  hhx[0] = hx[intx+1] = 0.
*/
NonEqui3D :: NonEqui3D
           (const int intx, const int inty, const int intz,
            const double xh[], const double yh[], const double zh[])
 : Grid3D(intx,inty,intz),
   hhx(new double[GetIntX()+2]), hhy(new double[GetIntY()+2]),
   hhz(new double[GetIntZ()+2]),
   weight_xl(new double [GetIntX()+2]), weight_xr(new double [GetIntX()+2]),
   weight_yl(new double [GetIntY()+2]), weight_yr(new double [GetIntY()+2]),
   weight_zl(new double [GetIntZ()+2]), weight_zr(new double [GetIntZ()+2])
{
//   hhx = new double[GetIntX()+2]; // Index 1,..,GetIntX() is usefull
//   hhy = new double[GetIntY()+2];
//   hhz = new double[GetIntZ()+2];
				// Interval lengths
  				// Values at Indices 0 and GetIntX()+1 are set to zero
  hhx[0] = hhx[GetIntX()+1] = 0.0;
  for (int i=1; i<=GetIntX(); i++)
  	hhx[i] = xh[i-1];

  hhy[0] = hhy[GetIntY()+1] = 0.0;
  for (int i=1; i<=GetIntY(); i++)
  	hhy[i] = yh[i-1];

  hhz[0] = hhz[GetIntZ()+1] = 0.0;
  for (int i=1; i<=GetIntZ(); i++)
  	hhz[i] = zh[i-1];

//		Interpolation weights
//		We assume that all even nodes (0...nx+1) on the fine grid
//		are also nodes on the coarse grid

//   weight_xl = new double [GetIntX()+2];
//   weight_xr = new double [GetIntX()+2];
//   weight_yl = new double [GetIntY()+2];
//   weight_yr = new double [GetIntY()+2];
//   weight_zl = new double [GetIntZ()+2];
//   weight_zr = new double [GetIntZ()+2];

  CalcIntWeights();

  return;
}

//---------------------------------------------------------------------------
NonEqui3D :: ~NonEqui3D()
{
  if (hhz!=0) delete [] hhz;
  if (hhy!=0) delete [] hhy;
  if (hhx!=0) delete [] hhx;
  if (weight_xl!=0) delete [] weight_xl;
  if (weight_xr!=0) delete [] weight_xr;
  if (weight_yl!=0) delete [] weight_yl;
  if (weight_yr!=0) delete [] weight_yr;
  if (weight_zl!=0) delete [] weight_zl;
  if (weight_zr!=0) delete [] weight_zr;
  return;
}

//---------------------------------------------------------------------------
/**	Copy constructor
							*/
NonEqui3D :: NonEqui3D(const NonEqui3D &orig)
           : Grid3D(orig)
{
  int i;

  hhx = new double[GetIntX()+2];
  hhy = new double[GetIntY()+2];
  hhz = new double[GetIntZ()+2];

  for (i=0; i<=GetIntX()+1; i++)
  	hhx[i] = orig.hhx[i];

  for (i=0; i<=GetIntY()+1; i++)
  	hhy[i] = orig.hhy[i];

  for (i=0; i<=GetIntZ()+1; i++)
  	hhz[i] = orig.hhz[i];

  weight_xl = new double [GetIntX()+2];
  weight_xr = new double [GetIntX()+2];
  weight_yl = new double [GetIntY()+2];
  weight_yr = new double [GetIntY()+2];
  weight_zl = new double [GetIntZ()+2];
  weight_zr = new double [GetIntZ()+2];

  for (i=0; i<=GetIntX()+1; i++)
    {
      weight_xl[i] = orig.weight_xl[i];
      weight_xr[i] = orig.weight_xr[i];
    }
  for (i=0; i<=GetIntY()+1; i++)
    {
      weight_yl[i] = orig.weight_yl[i];
      weight_yr[i] = orig.weight_yr[i];
    }
  for (i=0; i<=GetIntZ()+1; i++)
    {
      weight_zl[i] = orig.weight_zl[i];
      weight_zr[i] = orig.weight_zr[i];
    }

  return;
}

//---------------------------------------------------------------------------

NonEqui3D& NonEqui3D :: operator=(const NonEqui3D & orig)
{
 if ( this != &orig )
 {
   int i;
   Grid3D::operator=(orig);

   if (hhz!=0) delete [] hhz;
   if (hhy!=0) delete [] hhy;
   if (hhx!=0) delete [] hhx;

   hhx = new double[GetIntX()+2];
   hhy = new double[GetIntY()+2];
   hhz = new double[GetIntZ()+2];

   for (i=0; i<=GetIntX()+1; i++)
  	 hhx[i] = orig.hhx[i];

   for (i=0; i<=GetIntY()+1; i++)
  	 hhy[i] = orig.hhy[i];

   for (i=0; i<=GetIntZ()+1; i++)
  	 hhz[i] = orig.hhz[i];

   if (weight_xl!=0)  delete [] weight_xl;
   if (weight_xr!=0)  delete [] weight_xr;
   if (weight_yl!=0)  delete [] weight_yl;
   if (weight_yr!=0)  delete [] weight_yr;
   if (weight_zl!=0)  delete [] weight_zl;
   if (weight_zr!=0)  delete [] weight_zr;

   weight_xl = new double [GetIntX()+2];
   weight_xr = new double [GetIntX()+2];
   weight_yl = new double [GetIntY()+2];
   weight_yr = new double [GetIntY()+2];
   weight_zl = new double [GetIntZ()+2];
   weight_zr = new double [GetIntZ()+2];

   for (i=0; i<=GetIntX()+1; i++)
     {
       weight_xl[i] = orig.weight_xl[i];
       weight_xr[i] = orig.weight_xr[i];
     }
   for (i=0; i<=GetIntY()+1; i++)
     {
       weight_yl[i] = orig.weight_yl[i];
       weight_yr[i] = orig.weight_yr[i];
     }
   for (i=0; i<=GetIntZ()+1; i++)
     {
       weight_zl[i] = orig.weight_zl[i];
       weight_zr[i] = orig.weight_zr[i];
     }

 }
 return *this;
}

//---------------------------------------------------------------------------
void NonEqui3D :: CalcIntWeights()
{
  int i;

  weight_xl[0] = 0.0;
  for (i=1; i < GetIntX()+1; i++)
      weight_xl[i] = hx(i+1)/(hx(i)+hx(i+1));
  weight_xl[GetIntX()+1] = 0.0;

  for (i=0; i < GetIntX()+1; i++)
      weight_xr[i] = hx(i)/(hx(i)+hx(i+1));
  weight_xr[GetIntX()+1] = 0.0;

  weight_yl[0] = 0.0;
  for (i=1; i < GetIntY()+1; i++)
      weight_yl[i] = hy(i+1)/(hy(i)+hy(i+1));
  weight_yl[GetIntY()+1] = 0.0;

  for (i=0; i < GetIntY()+1; i++)
      weight_yr[i] = hy(i)/(hy(i)+hy(i+1));
  weight_yr[GetIntY()+1] = 0.0;

  weight_zl[0] = 0.0;
  for (i=1; i < GetIntZ()+1; i++)
      weight_zl[i] = hz(i+1)/(hz(i)+hz(i+1));
  weight_zl[GetIntZ()+1] = 0.0;

  for (i=0; i < GetIntZ()+1; i++)
      weight_zr[i] = hz(i)/(hz(i)+hz(i+1));
  weight_zr[GetIntZ()+1] = 0.0;

 return;
}

//---------------------------------------------------------------------------
void NonEqui3D :: GetCoords(double xcoord[]) const
{
 double x,y,z;
 unsigned int ijk = 0;

 z = 0.0;
 for (unsigned int k=0; k<=GetIntZ(); ++k)
 {
   y = 0.0;
   for (unsigned int j=0; j<=GetIntY(); ++j)
   {
     x = 0.0;
     for (unsigned int i=0; i<=GetIntX(); ++i, ++ijk)
     {
       xcoord[3*ijk+0] = x;
       xcoord[3*ijk+1] = y;
       xcoord[3*ijk+2] = z;
      x += hhx[i+1];
     }
     y += hhy[j+1];
   }
   z += hhz[k+1];
 }
 return;
}


//---------------------------------------------------------------------------

void NonEqui3D :: Interpolate(double *wh, const double *WH) const
{
 VDcopyC(VectorLength(), wh, 0.0);
 IntAdd(wh,WH);
}

//---------------------------------------------------------------------------

void NonEqui3D :: IntAdd(double *uh, const double *WH) const
{
  double 	t;
  unsigned int	nf1,nf2,nc1,nc2,ncx,ncy,ncz,i,j,k,ii,jj,kk,ijk,IJK;
  const NonEqui3D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  ncz = coarsegrid.GetIntZ();
  nf1 = Dim1();   nf2 = Dim2();
  nc1 = coarsegrid.Dim1();  nc2 = coarsegrid.Dim2();

  i = 2;
  for (ii=1; ii < ncz; ii++)
  {
    j = 2;
    for (jj=1; jj < ncy; jj++)
    {
      k = 2;
      ijk = idx(nf2,nf1, i ,j , k);
      IJK = idx(nc2,nc1, ii,jj, 1);
      for (kk=1; kk < ncx; kk++, IJK++)
      {
	t  = WH[IJK];

	uh[ijk]	+= t;

	uh[ijk-nf2]	+= wzr(i-1)*t;
	uh[ijk+nf2]	+= wzl(i+1)*t;
	uh[ijk-nf1]	+= wyr(j-1)*t;
	uh[ijk+nf1]	+= wyl(j+1)*t;
	uh[ijk-1]	+= wxr(k-1)*t;
	uh[ijk+1]	+= wxl(k+1)*t;

	uh[ijk-nf2-nf1]	+= wzr(i-1)*wyr(j-1)*t;
	uh[ijk-nf2+nf1]	+= wzr(i-1)*wyl(j+1)*t;
	uh[ijk+nf2-nf1]	+= wzl(i+1)*wyr(j-1)*t;
	uh[ijk+nf2+nf1]	+= wzl(i+1)*wyl(j+1)*t;
	uh[ijk-nf1-1]	+= wyr(j-1)*wxr(k-1)*t;
	uh[ijk-nf1+1]	+= wyr(j-1)*wxl(k+1)*t;
	uh[ijk+nf1-1]	+= wyl(j+1)*wxr(k-1)*t;
	uh[ijk+nf1+1]	+= wyl(j+1)*wxl(k+1)*t;
	uh[ijk-nf2-1]	+= wzr(i-1)*wxr(k-1)*t;
	uh[ijk-nf2+1]	+= wzr(i-1)*wxl(k+1)*t;
	uh[ijk+nf2-1]	+= wzl(i+1)*wxr(k-1)*t;
	uh[ijk+nf2+1]	+= wzl(i+1)*wxl(k+1)*t;

	uh[ijk-nf2-nf1-1] += wzr(i-1)*wyr(j-1)*wxr(k-1)*t;
	uh[ijk-nf2-nf1+1] += wzr(i-1)*wyr(j-1)*wxl(k+1)*t;
	uh[ijk-nf2+nf1-1] += wzr(i-1)*wyl(j+1)*wxr(k-1)*t;
	uh[ijk-nf2+nf1+1] += wzr(i-1)*wyl(j+1)*wxl(k+1)*t;
	uh[ijk+nf2-nf1-1] += wzl(i+1)*wyr(j-1)*wxr(k-1)*t;
	uh[ijk+nf2-nf1+1] += wzl(i+1)*wyr(j-1)*wxl(k+1)*t;
	uh[ijk+nf2+nf1-1] += wzl(i+1)*wyl(j+1)*wxr(k-1)*t;
	uh[ijk+nf2+nf1+1] += wzl(i+1)*wyl(j+1)*wxl(k+1)*t;

//    cout << k << " : " << wzr(k-1) << "  " << wzl(k+1) << endl;
        k += 2;  ijk += 2;
      }
      j += 2;
    }
    i += 2;
  }
}

//---------------------------------------------------------------------------

void NonEqui3D :: Restrict(const double *uh, double *UH) const
{
//   const double  CW = 1./8.;			//  normal weight
//   const double  CW = 1.0;			//  FEM weight

  double 	t,t2,t4,t8, tw;
  unsigned int	nf1,nf2,nc1,nc2,ncx,ncy,ncz,i,j,k,ii,jj,kk,ijk,IJK;
  const NonEqui3D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  ncz = coarsegrid.GetIntZ();
  nf1 = Dim1();   nf2 = Dim2();
  nc1 = coarsegrid.Dim1();  nc2 = coarsegrid.Dim2();

//  cout << " :::::::::::::::: " << nc1 << "  " << nc2 << endl;
  i = 2;
  for (ii=1; ii < ncz; ii++)
  {
    j = 2;
    for (jj=1; jj < ncy; jj++)
    {
      k = 2;
      ijk = idx(nf2,nf1, i ,j , k);
      IJK = idx(nc2,nc1, ii,jj, 1);
      for (kk=1; kk < ncx; kk++, IJK++)
      {

	tw = 1.0 +
             wzr(i-1) + wzl(i+1) +
	     wyr(j-1) + wyl(j+1) +
	     wxr(k-1) + wxl(k+1) +

             wzr(i-1)*( wyr(j-1) + wyl(j+1) ) +
	     wzl(i+1)*( wyr(j-1) + wyl(j+1) ) +
	     wyr(j-1)*( wxr(k-1) + wxl(k+1) ) +
	     wyl(j+1)*( wxr(k-1) + wxl(k+1) ) +
	     wzr(i-1)*( wxr(k-1) + wxl(k+1) ) +
	     wzl(i+1)*( wxr(k-1) + wxl(k+1) ) +

             wzr(i-1)*(
	         wyr(j-1)*(wxr(k-1) + wxl(k+1)) +
	         wyl(j+1)*(wxr(k-1) + wxl(k+1))
	              ) +
             wzl(i+1)*(
	         wyr(j-1)*(wxr(k-1) + wxl(k+1)) +
	         wyl(j+1)*(wxr(k-1) + wxl(k+1))
	             );

	t  = uh[ijk];

        t2 = wzr(i-1)*uh[ijk-nf2] + wzl(i+1)*uh[ijk+nf2] +
	     wyr(j-1)*uh[ijk-nf1] + wyl(j+1)*uh[ijk+nf1] +
	     wxr(k-1)*uh[ijk-  1] + wxl(k+1)*uh[ijk+  1] ;

        t4 = wzr(i-1)*( wyr(j-1)*uh[ijk-nf2-nf1] + wyl(j+1)*uh[ijk-nf2+nf1] ) +
	     wzl(i+1)*( wyr(j-1)*uh[ijk+nf2-nf1] + wyl(j+1)*uh[ijk+nf2+nf1] ) +
	     wyr(j-1)*( wxr(k-1)*uh[ijk-nf1-1]   + wxl(k+1)*uh[ijk-nf1+1]   ) +
	     wyl(j+1)*( wxr(k-1)*uh[ijk+nf1-1]   + wxl(k+1)*uh[ijk+nf1+1]   ) +
	     wzr(i-1)*( wxr(k-1)*uh[ijk-nf2-1]   + wxl(k+1)*uh[ijk-nf2+1]   ) +
	     wzl(i+1)*( wxr(k-1)*uh[ijk+nf2-1]   + wxl(k+1)*uh[ijk+nf2+1]   );

        t8 = wzr(i-1)*(
	         wyr(j-1)*(wxr(k-1)*uh[ijk-nf2-nf1-1] + wxl(k+1)*uh[ijk-nf2-nf1+1]) +
	         wyl(j+1)*(wxr(k-1)*uh[ijk-nf2+nf1-1] + wxl(k+1)*uh[ijk-nf2+nf1+1])
	              ) +
             wzl(i+1)*(
	         wyr(j-1)*(wxr(k-1)*uh[ijk+nf2-nf1-1] + wxl(k+1)*uh[ijk+nf2-nf1+1]) +
	         wyl(j+1)*(wxr(k-1)*uh[ijk+nf2+nf1-1] + wxl(k+1)*uh[ijk+nf2+nf1+1])
	             );

//	cout << IJK << "(" << ii << "," << jj << "," << kk << ") = "
//		<< tw << endl;
//	cout << t << "  " << t2 << "  " << t4 << "  " << t8 << endl;

//	UH[IJK] = (t+t2+t4+t8) * CW ;
	UH[IJK] = (t+t2+t4+t8) / tw ;     // *2 bei FEM

	k += 2;  ijk += 2;
      }
      j += 2;
    }
    i += 2;
  }
//   CheckBoundaryZero(uh);
//   cout << "--------------------" << endl;
//   coarsegrid.CheckBoundaryZero(UH);
}

//---------------------------------------------------------------------------
//			h[0:n+1], xhf[0:2*n-1]

// void NonEqui3D :: Refine1D(const int n, const double h[], double xhf[]) const
// {
//   int i,ii;
//   double tr,tl;
//
//   i = 0;
//   tl = h[1]; tr = h[2];
//   xhf[i++] = tl/(tl+tr)*h[1];
//   xhf[i++] = tr/(tl+tr)*h[1];
//   for (ii=2; ii<n; ii++)
//    {
//      tl = h[ii-1]+h[ii];
//      tr = h[ii]+h[ii+1];
//      xhf[i++] = tl/(tl+tr)*h[ii];
//      xhf[i++] = tr/(tl+tr)*h[ii];
//    }
//   tl = h[n-1];
//   tr = h[n];
//   xhf[i++] = tl/(tl+tr)*h[n];
//   xhf[i++] = tr/(tl+tr)*h[n];
//
//   return;
// }

void NonEqui3D :: Refine1D(const int n, const double h[], double xhf[]) const
{
  int i,ii;
  double tr,tl;

  i = 0;
  tl = h[1]; tr = h[2];
  xhf[i++] = 0.5*h[1];
  xhf[i++] = 0.5*h[1];
  for (ii=2; ii<n; ii++)
   {
     tl = h[ii-1]+h[ii];
     tr = h[ii]+h[ii+1];
     xhf[i++] = 0.5*h[ii];
     xhf[i++] = 0.5*h[ii];
   }
  tl = h[n-1];
  tr = h[n];
  xhf[i++] = 0.5*h[n];
  xhf[i++] = 0.5*h[n];

  return;
}
//---------------------------------------------------------------------------

void NonEqui3D :: Refine(NonEqui3D & fine) const
{
  int nxc = GetIntX(), nyc =  GetIntY(), nzc = GetIntZ();
  int nxf = 2*nxc,     nyf = 2*nyc,      nzf = 2*nzc;
  double *xhf, *yhf, *zhf;

  xhf = new double [nxf];
  yhf = new double [nyf];
  zhf = new double [nzf];

  Refine1D(nxc, hhx, xhf);
  Refine1D(nyc, hhy, yhf);
  Refine1D(nzc, hhz, zhf);

  fine = NonEqui3D(nxf,nyf,nzf, xhf,yhf,zhf);
  fine.coarsegrid_ = this;

  delete [] zhf; delete [] yhf; delete [] xhf;
}

//---------------------------------------------------------------------------
//			h[0:2*n+1], xhc[0:n-1]

void NonEqui3D :: Coarse1D(const int n, const double h[], double xhf[]) const
{
  int i;

  for (i=0; i<n; i++)
   {
     xhf[i] = h[2*i+1] + h[2*i+2];
   }

  return;
}

//---------------------------------------------------------------------------

bool NonEqui3D :: Coarse(NonEqui3D & coarse)
{
  int nxf = GetIntX(), nyf =  GetIntY(), nzf = GetIntZ();
  int nxc = nxf/2,     nyc = nyf/2,      nzc = nzf/2;
  double *xhc, *yhc, *zhc;
  bool bcoarse;

  bcoarse = (2*nxc==nxf && nxc>2) && (2*nyc==nyf && nyc>2)
                                  && (2*nzc==nzf && nzc>2);
  if (bcoarse)	// i.e., all coarse dimensions are even
   {

    xhc = new double [nxc];
    yhc = new double [nyc];
    zhc = new double [nzc];

    Coarse1D(nxc, hhx, xhc);
    Coarse1D(nyc, hhy, yhc);
    Coarse1D(nzc, hhz, zhc);

    coarse = NonEqui3D(nxc,nyc,nzc, xhc,yhc,zhc);
    this->coarsegrid_ = &coarse;

    delete [] zhc; delete [] yhc; delete [] xhc;

//	Calculate Interpolation weights!!
    CalcIntWeights();

#ifdef DEBUG
   cout << endl << nxf << "  " << nxc << endl;
   cout << endl << "HHX : " ;
   for (int i=0; i<nxc+2; i++)
  	  cout << coarse.hhx[i] << "  ";
    cout << endl;
#endif
   }
  else
   {
    this->coarsegrid_ = 0;
   }

  return bcoarse;
}
