#include <iostream>
#include <cstdlib>
#include <cmath>
using namespace std;
#include "vector.hpp"
#include "grid.hpp"
#include "grid2d.hpp"
#include "laplace2d.hpp"

//#define DEBUG_MATRIX

const     int NAUX=8193;
static double tt[NAUX];		// aux. array for ADI  (FIXXXIT)
//static double  rr[NAUX];

//	###################################################################
//			DiagLaplaceEqui2D
//	###################################################################
//	Laplace 2D, equidistant tensor product grid

//---------------------------------------------------------------------------

DiagLaplaceEqui2D ::
DiagLaplaceEqui2D(const int mx, const int my, const bool coarse)
 : Equi2D(mx,my), bx(0.0), by(0.0), diag(0), diagx(0), diagy(0), offx(0), offy(0)
{
  InitOperator(coarse);
}

//---------------------------------------------------------------------------

DiagLaplaceEqui2D :: DiagLaplaceEqui2D(const int mx, const bool coarse)
 : Equi2D(mx), bx(0.0), by(0.0), diag(0), diagx(0), diagy(0), offx(0), offy(0)
{
  InitOperator(coarse);
}

//---------------------------------------------------------------------------

DiagLaplaceEqui2D :: ~DiagLaplaceEqui2D()
{
 DeleteGridVector(diag);
//  cout << diagx << "  " << offy << "  " << offy - diagx << endl;
 if (diagx!=0)
 {
   delete [] diagx;  delete [] diagy;
   delete [] offx;   delete [] offy;
 }
}

//---------------------------------------------------------------------------

DiagLaplaceEqui2D :: DiagLaplaceEqui2D(const DiagLaplaceEqui2D &orig)
: Equi2D(orig), bx(orig.bx), by(orig.by),
  diag(0), diagx(0), diagy(0), offx(0), offy(0)
{
 NewGridVector(diag);
 VDcopy( orig.VectorLength(), diag, 1, orig.diag, 1 );
 FactorADIMatrices();
}

//---------------------------------------------------------------------------
DiagLaplaceEqui2D& DiagLaplaceEqui2D :: operator=(const DiagLaplaceEqui2D & orig)
{
 if ( this != &orig )
 {
   Equi2D::operator=(orig);

   bx = orig.bx;  by = orig.by;
   DeleteGridVector(diag);
   NewGridVector(diag);
   VDcopy( orig.VectorLength(), diag, 1, orig.diag, 1 );
   FactorADIMatrices();
 }
 return *this;
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui2D ::InitOperator(const bool coarse, const double *cc)
{
//   int n=VectorLength();

  bx = -1.0/hx/hx;  by = -1.0/hy/hy;

  DeriveDiag(cc);	//	bx,by are used in diag calculation
//   if (diag==0)   NewGridVector(diag);
//   VDcopyC( n, diag, -2.0*(bx+by) );
//   if  (cc !=0 ) 	//			Add Helmholtz  term
//   	VDadd( n, diag, cc );
//   SetBoundaryConst(diag, 1.0);		//	Dirichlet B.C.
  if (coarse)
    {			//	diag is used in matrix calculation
      DeriveMatrix();
#ifdef DEBUG_MATRIX
      cout << endl << "Coarse Matrix" << endl;
      PrintMatrix(VectorLength(), mat);
#endif
      InvertMatrix();
#ifdef DEBUG_MATRIX
      cout << endl << "Coarse Matrix" << endl;
      PrintMatrix(VectorLength(), mat);
#endif
    }
 FactorADIMatrices();
}
//---------------------------------------------------------------------------

void DiagLaplaceEqui2D :: DeriveDiag(const double *cc)
{
 int n=VectorLength();

 if (diag==0)   NewGridVector(diag);
 VDcopyC( n, diag, -2.0*(bx+by) );
 if  (cc !=0 ) 			//	Add Helmholtz  term
       VDadd( n, diag, cc );
 SetBoundaryConst(diag, 1.0); 	//	Dirichlet boundary

#ifdef DEBUG_MATRIX
  cout << endl << "Diagonal in DiagLaplaceEqui2D" << endl;
  PrintVector(n,diag);
#endif
}
//---------------------------------------------------------------------------

void DiagLaplaceEqui2D :: Refine(DiagLaplaceEqui2D & fine) const
{
//						not a coarse grid
  fine = DiagLaplaceEqui2D(2*nx,2*ny,false);
  fine.coarsegrid_ = this;
//  fine.DeriveMatrix()  must not be executed since this is already a fine grid ;
}

//---------------------------------------------------------------------------
//	Derive matrix for the 7 point stencil
//		and Dirichlet boundary conditions

void DiagLaplaceEqui2D :: DeriveMatrix()
{
 int i,j,k,n1,nfx,nfy,ijk,id;
 int nn = VectorLength();
 unsigned long int nn2 = nn*nn;

 n1  = Dim1();
 nfx = GetIntX();
 nfy = GetIntY();

//	Allocate memory for fully occupied matrix if called the first time
//			(Assumes no changes in dimensions!!)
 if (mat==0)   mat = new double[nn2];
 VDcopyC(nn2, mat, 0.0);			// set matrix to Zero
 for ( i=0; i<nn; ++i )  mat[i*(nn+1)] = 1.0;	// Initialize diagonal with 1

 for ( j=1; j<nfy; ++j )
  {
   for ( k=1; k<nfx; ++k )
    {
      ijk = idx(j, k);
      id  = ijk*nn + ijk;
      mat[id] = diag[ijk];
      if ( k != 1    )  mat[id-1 ] = bx;
      if ( k != nfx-1 ) mat[id+1 ] = bx;
      if ( j != 1    )  mat[id-n1] = by;
      if ( j != nfy-1 ) mat[id+n1] = by;
   }
  }
#ifdef DEBUG_MATRIX
  cout << endl << "Coarse Matrix in 2D" << endl;
  PrintMatrix(nn,mat);
#endif
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui2D :: DirectSolve(double *u, const double *f) const
{
 if ( nx != 2 ||  ny != 2)
  {
   Equi2D::DirectSolve(u,f);
  }
 else
  {
   u[4] = f[4] / diag[4];
 }
}
//---------------------------------------------------------------------------
// faster  (1.41 sec --> 1.14 sec.
//  	unsauber bzgl. letzer Eintrag in Zeile und gesamter letzter Zeile
//	Zeilen 218--220  (und 213--214 ??)
//	Unsauberkeit behoben. (// GH_N), code ist evtl. 0.1% langsamer

void DiagLaplaceEqui2D :: DefRes(const double *fh, const double *uh, double *DH) const
{
//  const double  CW = 1.;			//  4 because h^2 -> rhs
  const double  CW = 1./4.;			//  normal weight

  double 	t1,t2b,t2c,t4c, u1,u2b,u2c,u4c,
  		u1_old, u2b_old, u2c_new, u4c_new;
  unsigned int	nf1,nc1,ncx,ncy,j,k,jj,kk,ijk,IJK;
#ifdef REMOVE_MEM_BUG
          bool  bj,bk;				// GH_N
#endif
  const DiagLaplaceEqui2D   &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  nf1 = Dim1();   nc1 = coarsegrid.Dim1();

  for (IJK=0; IJK<=ncx; ++IJK)  DH[IJK] = 0.0;	// init sum
  j = 2;
  for (jj=1; jj <= ncy; ++jj)
   {
#ifdef REMOVE_MEM_BUG
     bj = (j!=ny);				// GH_N
#endif
     u2b = uh[idx(j-1,0)];  u4c_new = uh[idx(j-1,1)];
     u1  = uh[idx(j  ,0)];  u2c_new = uh[idx(j  ,1)];
     k = 2;
     IJK = IDX(jj, 1);
     DH[IJK-1] = 0.0;				// init sum
     for (kk=1; kk <= ncx; ++kk, ++IJK)
     {
#ifdef REMOVE_MEM_BUG
       bk = (k!=nx);				// GH_N
#endif
       u2b_old = u2b;   u4c = u4c_new;
       u1_old =  u1;   u2c = u2c_new;

       ijk = idx(j-1,k-1);
       t4c  = fh[ijk]   - by * uh[ijk-nf1];
       t2b  = fh[ijk+1] - by * uh[ijk-nf1+1];
       u2b     = uh[ijk+1];
#ifdef REMOVE_MEM_BUG
       u4c_new = bk ? uh[ijk+2] : 0.0;				// GH_N
#else
       u4c_new = uh[ijk+2];
#endif
//			Klammern bzw. Nicht-Klammern sind optimal fr SGI
//					(R�selhaft)
       t4c -= diag[ijk]   * u4c + bx * u2b  + bx * u2b_old;
       t2b -= diag[ijk+1] * u2b + bx * (u4c  + u4c_new);

       ijk += nf1;

       t2c  = fh[ijk] - bx * u1_old;
       u2c  = uh[ijk];
       u1   = uh[ijk+1];
#ifdef REMOVE_MEM_BUG
       u2c_new = bk ? uh[ijk+2] : 0.0;		// GH_N
#else
       u2c_new = uh[ijk+2];
#endif
       t4c -= by * u2c;
       t2b -= by * u1;
       t2c -= diag[ijk] * u2c + bx * u1;
       t1   = fh[ijk+1] - diag[ijk+1] * u1 - bx * (u2c + u2c_new);
#ifdef REMOVE_MEM_BUG
       t2c -= by * ( (bj ? uh[ijk+nf1] : 0.0) +u4c);		// GH_N
       t1  -= by * ( (bj ? uh[ijk+nf1+1] : 0.0) +u2b);		// GH_N
#else
       t2c -= by * ( uh[ijk+nf1]   +u4c);
       t1  -= by * ( uh[ijk+nf1+1] +u2b);
#endif

       t4c *= CW/4.0;
       t2c *= CW/2.0;
       t2b *= CW/2.0;
       t1  *= CW;

       DH[IJK-nc1-1]     += t4c;
       DH[IJK-nc1]       += t4c+t2b;
       DH[IJK-1]         += t4c+t2c;
       DH[IJK]            = t4c+t2b+t2c+t1;

       k += 2;
     }
     DH[IDX(jj, 0)] = DH[IDX(jj, ncx)] = 0.0;	// Dirichlet
     j += 2;
   }
  for (kk=0; kk<=ncx; kk++)
       DH[kk] = DH[IDX(ncy,0)+kk] = 0.0; 	// Dirichlet

  return;
}

//---------------------------------------------------------------------------
// (1.41 sec --> 1.29 sec.)
// void DiagLaplaceEqui2D :: DefRes(const double *fh, const double *uh, double *DH) const
// {
//   const double  CW = 1./4.;			//  normal weight
//
//   double 	t1,t2b,t2c,t4c, u1,u2b,u2c,u4c;
//   unsigned int	nf1,nc1,ncx,ncy,j,k,jj,kk,ijk,IJK;
//   const DiagLaplaceEqui2D   &coarsegrid = CoarseGrid();
//
//   ncx = coarsegrid.GetIntX();
//   ncy = coarsegrid.GetIntY();
//   nf1 = Dim1();   nc1 = coarsegrid.Dim1();
//
//   for (IJK=0; IJK<=ncx; IJK++)  DH[IJK] = 0.0;	// init sum
//   j = 2;
//   for (jj=1; jj <= ncy; jj++)
//    {
//      k = 2;
//      IJK = IDX(jj, 1);
//      DH[IJK-1] = 0.0;				// init sum
//      for (kk=1; kk <= ncx; kk++, IJK++)
//      {
//        ijk = idx(j-1,k-1);
//
//        t4c  = fh[ijk]   - by * uh[ijk-nf1];
//        t2b  = fh[ijk+1] - by * uh[ijk-nf1+1];
//        t4c -= bx * uh[ijk-1];
//        u4c  = uh[ijk];
//        u2b  = uh[ijk+1];
//        t4c -= diag[ijk] * u4c + bx * u2b;
//        t2b -= bx * (u4c + uh[ijk+2]) + diag[ijk+1] * u2b;
//
//        ijk += nf1;
//
//        t2c  = fh[ijk] - bx * uh[ijk-1];
//        u2c  = uh[ijk];
//        u1   = uh[ijk+1];
//        t4c -= by * u2c;
//        t2b -= by * u1;
//        t2c -= diag[ijk] * u2c + bx * u1;
//        t1   = fh[ijk+1] - diag[ijk+1] * u1 - bx * (u2c + uh[ijk+2]);
//        t2c -= by * (uh[ijk+nf1]+u4c);
//        t1  -= by * (uh[ijk+nf1+1]+u2b);
//
//        t4c *= CW/4.0;
//        t2c *= CW/2.0;
//        t2b *= CW/2.0;
//        t1  *= CW;
//
//        DH[IJK-nc1-1]     += t4c;
//        DH[IJK-nc1]       += t4c+t2b;
//        DH[IJK-1]         += t4c+t2c;
//        DH[IJK]            = t4c+t2b+t2c+t1;
//
//        k += 2;
//      }
//      DH[IDX(jj, 0)] = DH[IDX(jj, ncx)] = 0.0;	// Dirichlet
//      j += 2;
//    }
//   for (kk=0; kk<=ncx; kk++)
//        DH[kk] = DH[IDX(ncy,0)+kk] = 0.0; 	// Dirichlet
//
//   return;
// }

//---------------------------------------------------------------------------
//
// void DiagLaplaceEqui2D :: DefRes(const double *fh, const double *uh, double *DH) const
// {
// //  const double  CW = 1.;			//  4 because h^2 -> rhs
//   const double  CW = 1./4.;			//  normal weight
//
//   double 	t,t1,t2b,t2c,t4c, u1,u2b,u2c,u4c;
//   unsigned int	nf1,nc1,ncx,ncy,j,k,jj,kk,ijk,IJK;
//   const DiagLaplaceEqui2D   &coarsegrid = CoarseGrid();
//
//   ncx = coarsegrid.GetIntX();
//   ncy = coarsegrid.GetIntY();
//   nf1 = Dim1();   nc1 = coarsegrid.Dim1();
//
// //  VDcopyC(coarsegrid.VectorLength(), DH, 0.0e0);
//
//   for (IJK=0; IJK<=ncx; IJK++)  DH[IJK] = 0.0;	// init sum
//   j = 2;
//   for (jj=1; jj <= ncy; jj++)
//    {
//      k = 2;
//      IJK = IDX(jj, 1);
//      DH[IJK-1] = 0.0;				// init sum
//      for (kk=1; kk <= ncx; kk++, IJK++)
//      {
// //			Compiler besser untersttzen!!
//        ijk = idx(j-1,k-1);
//        u4c = uh[ijk];		t4c = fh[ijk];
//        u2b = uh[ijk+1];		t2b = fh[ijk+1];
//        u2c = uh[ijk+nf1];	t2c = fh[ijk+nf1];
//        u1  = uh[ijk+nf1+1]; 	t1  = fh[ijk+nf1+1];
//
//        t   = by*(uh[ijk-nf1] + u2c) +
// 	     bx*(uh[ijk-1]   + u2b);
// //       t4c = ( fh[ijk] - diag[ijk] * u4c - t ) / 4.0 * CW;
//        t4c = ( t4c - diag[ijk] * u4c - t ) / 4.0 * CW;
//
// //       ijk = idx(j-1,k);
//        ijk++;
//        t   = by*(uh[ijk-nf1] + u1) +
// 	     bx*(u4c   + uh[ijk+1]);	   //???
// //       t2b = ( fh[ijk] - diag[ijk] * u2b - t ) / 2.0 * CW;
//        t2b = ( t2b - diag[ijk] * u2b - t ) / 2.0 * CW;
// //h	if (k==nx)  t2b=0.0;		// inefficient
//
// //       ijk = idx(j,k-1);
//        ijk += nf1-1;
//        t   = by*(u4c + uh[ijk+nf1]) +
// 	     bx*(uh[ijk-1]   + u1);
// //       t2c = ( fh[ijk] - diag[ijk] * u2c - t ) / 2.0 * CW;
//        t2c = ( t2c - diag[ijk] * u2c - t ) / 2.0 * CW;
// //h	if (j==ny)  t2c=0.0;		// inefficient
//
// //       ijk = idx(j,k);
//        ijk++;
//        t   = by*(u2b + uh[ijk+nf1]) +
// 	     bx*(u2c   + uh[ijk+1]);
// //       t1  = ( fh[ijk] - diag[ijk] * u1 - t ) * CW;
//        t1  = ( t1 - diag[ijk] * u1 - t ) * CW;
// //h	if (j==ny || k==nx)  t1=0.0;		// inefficient
//
//        DH[IJK-nc1-1]     += t4c;
//        DH[IJK-nc1]       += t4c+t2b;
//        DH[IJK-1]         += t4c+t2c;
// //       DH[IJK]           += t4c+t2b+t2c+t1;
//        DH[IJK]            = t4c+t2b+t2c+t1;
//
//        k += 2;
//      }
//      DH[IDX(jj, 0)] = DH[IDX(jj, ncx)] = 0.0;	// Dirichlet
//      j += 2;
//    }
//    for (kk=0; kk<=ncx; kk++)
//      DH[kk] = DH[IDX(ncy,0)+kk] = 0.0; 	// Dirichlet
//
//
// //  CheckBoundaryZero(uh);			// OK
// //  coarsegrid.CheckBoundaryZero(DH);	// boundary is not zero !!
// //					Set boundary to Zero !!
// //DD  coarsegrid.SetBoundaryZero(DH);
// }

//---------------------------------------------------------------------------

// void DiagLaplaceEqui2D :: DefRes(const double *fh, const double *uh, double *DH) const
// {
// //  const double  CW = 1.;			//  4 because h^2 -> rhs
//   const double  CW = 1./4.;			//  normal weight
//
//   double 	t,t1,t2b,t2c,t4c;
//   unsigned int	nf1,nc1,ncx,ncy,j,k,jj,kk,ijk,IJK;
//   const DiagLaplaceEqui2D   &coarsegrid = CoarseGrid();
//
//   ncx = coarsegrid.GetIntX();
//   ncy = coarsegrid.GetIntY();
//   nf1 = Dim1();   nc1 = coarsegrid.Dim1();
//
//   VDcopyC(coarsegrid.VectorLength(), DH, 0.0e0);
//
//   j = 2;
//   for (jj=1; jj <= ncy; jj++)
//    {
//      k = 2;
//      IJK = IDX(jj, 1);
//      for (kk=1; kk <= ncx; kk++, IJK++)
//      {
// //			Compiler besser untersttzen!!
//        ijk = idx(j-1,k-1);
//        t   = by*(uh[ijk-nf1] + uh[ijk+nf1]) +
// 	     bx*(uh[ijk-1]   + uh[ijk+1]);
//        t4c = ( fh[ijk] - diag[ijk] * uh[ijk] - t ) / 4.0 * CW;
//
//        ijk = idx(j-1,k);
//        t   = by*(uh[ijk-nf1] + uh[ijk+nf1]) +
// 	     bx*(uh[ijk-1]   + uh[ijk+1]);
//        t2b = ( fh[ijk] - diag[ijk] * uh[ijk] - t ) / 2.0 * CW;
// //h	if (k==nx)  t2b=0.0;		// inefficient
//
//        ijk = idx(j,k-1);
//        t   = by*(uh[ijk-nf1] + uh[ijk+nf1]) +
// 	     bx*(uh[ijk-1]   + uh[ijk+1]);
//        t2c = ( fh[ijk] - diag[ijk] * uh[ijk] - t ) / 2.0 * CW;
// //h	if (j==ny)  t2c=0.0;		// inefficient
//
//        ijk = idx(j,k);
//        t   = by*(uh[ijk-nf1] + uh[ijk+nf1]) +
// 	     bx*(uh[ijk-1]   + uh[ijk+1]);
//        t1  = ( fh[ijk] - diag[ijk] * uh[ijk] - t ) * CW;
// //h	if (j==ny || k==nx)  t1=0.0;		// inefficient
//
//        DH[IJK-nc1-1]     += t4c;
//        DH[IJK-nc1]       += t4c+t2b;
//        DH[IJK-1]         += t4c+t2c;
//        DH[IJK]           += t4c+t2b+t2c+t1;
//
//        k += 2;
//      }
//      j += 2;
//    }
//
// //  CheckBoundaryZero(uh);			// OK
// //  coarsegrid.CheckBoundaryZero(DH);	// boundary is not zero !!
// //					Set boundary to Zero !!
//   coarsegrid.SetBoundaryZero(DH);
// }

//---------------------------------------------------------------------------

void DiagLaplaceEqui2D :: Mult(double *f, const double *u) const
{
 double       t;
 unsigned int n1 = Dim1();

 for (unsigned int  k=0;   k<= GetIntX();  ++k)  f[k] = u[k];
 for (unsigned int  j=1;   j < GetIntY(); ++j)
 {
   unsigned int jj = idx(j,1);
   f[jj-1] = u[jj-1];		// Dirichlet
   for (unsigned int k=1; k < GetIntX(); ++k, ++jj)
   {
     t = by*(u[jj-n1] + u[jj+n1]) +
	 bx*(u[jj- 1] + u[jj +1]);

     f[jj] = diag[jj] * u[jj] + t;
   }
   f[jj] = u[jj];		// Dirichlet
 }
 for (unsigned int k=idx(GetIntY(),0);  k<=idx(GetIntY(),GetIntX());  ++k)
 	f[k] = u[k];
 return;
}

//---------------------------------------------------------------------------
void DiagLaplaceEqui2D ::
Defect(const double *f, const double *u, double *def) const
{
 double 	t;
 unsigned int	j,n1,jj, ja, je;

 SetBoundaryZero(def);			// homogeneous Dirichlet B.C.

 n1 = Dim1();

 ja = idx(1,0);
 for (j=1; j < GetIntY(); ++j)
 {
   je = ja+n1;
//   for (k=1; k < GetIntX(); ++k, ++jj)
   for (jj = ja+1; jj < je-1; ++jj)
   {
     t = by*(u[jj-n1] + u[jj+n1]) +
	 bx*(u[jj- 1] + u[jj +1]);

     def[jj] = f[jj] - diag[jj] * u[jj] - t;
   }
   ja = je;
 }

 return;
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui2D :: Jacobi(double *u, const double *f, double *def,
  			const double omega, const int iter) const
{
 int n = VectorLength();

 Defect(f,u,def);
 VDdiv(n,def,diag);
 VDXpaY(n,u,omega,def);
}

//---------------------------------------------------------------------------
//	use of aux. variable  tm  accelerates function by 10%
//					(1.98 -> 1.80 sec)

void DiagLaplaceEqui2D ::
GaussSeidelForward(double *u, const double *f, const int iter) const
{
  double 	t,tb;
  unsigned int	j,k,n1,jj;

  n1 = Dim1();

  for (j=1; j < GetIntY(); ++j)
  {
    jj = idx(j,1);
    tb = u[jj-1];
    for (k=1; k < GetIntX(); ++k, ++jj)
    {
      t = by*(u[jj-n1] + u[jj+n1]) +
	  bx*(tb    + u[jj+1] );

      u[jj]= (f[jj] - t) / diag[jj];
      tb = u[jj];
    }
  }

#ifdef DEBUG
 CheckBoundaryZero(u);
#endif
}

//---------------------------------------------------------------------------

// void DiagLaplaceEqui2D ::
// GaussSeidelForward(double *u, const double *f, const int iter) const
// {
//   double 	t;
//   unsigned int	j,k,n1,jj;
//
//   n1 = Dim1();
//
//   for (j=1; j < GetIntY(); j++)
//   {
//     jj = idx(j,1);
//     for (k=1; k < GetIntX(); k++, jj++)
//     {
//       t = by*(u[jj-n1] + u[jj+n1]) +
// 	  bx*(u[jj-1]  + u[jj+1] );
//
//       u[jj]= (f[jj] - t) / diag[jj];
//     }
//   }
// #ifdef DEBUG
//  CheckBoundaryZero(u);
// #endif
// }
//---------------------------------------------------------------------------

void DiagLaplaceEqui2D ::
GaussSeidelBackward(double *u, const double *f, const int iter) const
{
  double 	t;
  unsigned int	j,k,n1,jj;

  n1 = Dim1();

  for (j=GetIntY()-1; j > 0; --j)
  {
    jj = idx(j,GetIntX()-1);
    for (k=GetIntX()-1; k > 0; --k, --jj)
    {
      t = by*(u[jj-n1] + u[jj+n1]) +
	  bx*(u[jj-1]  + u[jj+1]);

      u[jj]= (f[jj] - t) / diag[jj];
    }
  }
}

//---------------------------------------------------------------------------
//	slightly faster by lines commented as //sf
//		by 10% (0.84 --> 0.74 sec.)

double DiagLaplaceEqui2D :: DefectNorm(const double *u, const double *f) const
{
  double 	t,s,r,tl,tm,tr;
  unsigned int	j,k,n1,jj;

  n1 = Dim1();
  r  = 0.0;

  for (j=1; j < GetIntY(); ++j)
  {
    jj = idx(j,1);

    tm = u[jj-1]; tr = u[jj];				// sf
    for (k=1; k < GetIntX(); ++k, ++jj)
    {
      tl = tm; tm = tr; tr = u[jj+1]; 			// sf
//      t = by*(u[jj-n1] + u[jj+n1]) +
//	  bx*(u[jj-1]  + u[jj+1] );
//      s = f[jj] - diag[jj] * u[jj] - t;

      t = by*(u[jj-n1] + u[jj+n1]) + bx*(tl+tr);	// sf
      s = f[jj] - diag[jj] *tm - t;			// sf
      r += s*s;
    }
  }
  return(sqrt(r));
 }

//---------------------------------------------------------------------------

// double DiagLaplaceEqui2D :: DefectNorm(const double *u, const double *f) const
// {
//   double 	t,s,r;
//   unsigned int	j,k,n1,jj;
//
//   n1 = Dim1();
//
//   r = 0.0;
//
//   for (j=1; j < GetIntY(); j++)
//   {
//     jj = idx(j,1);
//     for (k=1; k < GetIntX(); k++, jj++)
//     {
//       t = by*(u[jj-n1] + u[jj+n1]) +
// 	  bx*(u[jj-1]  + u[jj+1] );
//
//       s = f[jj] - diag[jj] * u[jj] - t;
//       r += s*s;
//     }
//   }
//
//   return(sqrt(r));
//  }

//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., Laplace + c
//	rho*I = sqrt(2)*Laplace_y + c/2 *I

void DiagLaplaceEqui2D :: ADIx(double u[], const double f[]) const
{
  const double CSQ2 = 1.0-sqrt(2.0);
  const unsigned int n1 = Dim1();
  const unsigned int nfx = GetIntX(),  nfy = GetIntY();
//old  const double rho = sqrt(8.0)*nfy*nfy;
        double tm;
  unsigned int i,j,jj;

  for (i=0; i <= nfx; i++)   u[i] = 0.0;
  for (i=0; i <= nfx; i++)  tt[i] = 0.0;
  for (j=1; j < nfy; j++)
  {
   jj = idx(j,0);
   tm = u[jj];			//	right hand side
   u[jj] = 0.0;				// Dirichlet
   for (i=1; i < nfx; i++)
     {					// f + (rho-D_y)tm ...
       tm   = u[jj+i];
//old       u[jj+i] = f[jj+i] + (rho+2*by)*tm - by*(u[jj+n1+i]+tt[i]);
//       u[jj+i] = f[jj+i] + 2*CSQ2*by*tm - by*(u[jj+n1+i]+tt[i]);
       u[jj+i] = f[jj+i] + by*(2*CSQ2*tm - u[jj+n1+i]-tt[i]);
       tt[i] = tm;
     }
   u[jj+nfx] = 0.0;

   SubstFactorSymmTriDiag(nfx+1, diagx, offx, u+jj);
  }
  for (i=idx(nfy,0); i <= idx(nfy,nfx); i++)  u[i] = 0.0;

  return;
}
//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C.,  Laplace + c
//	rho*I = sqrt(2)*Laplace_x + c/2 *I

void DiagLaplaceEqui2D :: ADIy(double u[], const double f[]) const
{
  const double CSQ2 = 1.0-sqrt(2.0);
  const unsigned int n1 = Dim1();
  const unsigned int nfx = GetIntX(),  nfy = GetIntY();
//old  const double rho = sqrt(8.0)*nfx*nfx;
        double tm,tb;
   unsigned int i,j,jj;

  for (i=0; i <= nfx; i++)  u[i] = 0.0;
  for (j=1; j < nfy; j++)
  {
    jj = idx(j,0);		//	right hand side
    tm = u[jj];
    u[jj] = 0.0;			// Dirichlet
    for (i=1; i < nfx; i++)
     {					// f + (rho-D_x)tm ...
       tb = tm;
       tm = u[jj+i];
//old       u[jj+i] = f[jj+i] + (rho+2*bx)*tm - bx*(u[jj+i+1]+tb);
//       u[jj+i] = f[jj+i] + 2*CSQ2*bx*tm - bx*(u[jj+i+1]+tb);
       u[jj+i] = f[jj+i] + bx*(2*CSQ2*tm - u[jj+i+1]-tb);
     }
    u[jj+nfx] = 0.0;			// Dirichlet
  }
  for (i=idx(nfy,0); i <= idx(nfy,nfx); i++)  u[i] = 0.0;

  SubstFactorSymmTriDiag_MultT(nfy+1, diagy, offy, u, n1);

  return;
}

//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., only Laplace
//	Factorization (rho !!) is done in FactorADIMatrices()

void DiagLaplaceEqui2D ::
ADIForward(double *u, const double *f, double *aux, const int iter) const
{
#ifdef DEBUG
  cout << "#########  u_in  ###########" << endl;
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
  cout << "#########  f  ###########" << endl;
  PrintVector(VectorLength(),f);
  CheckBoundaryZero(f);
#endif

  ADIx(u,f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
#endif

  ADIy(u,f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
#endif
  return;
}
//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., only Laplace
//	Factorization (rho !!) is done in FactorADIMatrices()

void DiagLaplaceEqui2D ::
ADIBackward(double *u, const double *f, double *aux, const int iter) const
{
 ADIy(u,f);
 ADIx(u,f);
#ifdef DEBUG
 CheckBoundaryZero(u);
#endif
  return;
}

//---------------------------------------------------------------------------
void DiagLaplaceEqui2D :: FactorADIMatrices()
{
 const double CSQ2 = 1.0-sqrt(2.0);
 double       diagsum;
 int          i,j, nfx=GetIntX(), nfy=GetIntY();

//   In order to use ADI, the lines
//     laplace2d.cc:58: FactorADIMatrices();
//     laplace2d.cc:72:   FactorADIMatrices();
//     laplace2d.cc:104: FactorADIMatrices();
//  call this routine and no_adi has to be set "true.
bool no_adi=true;
if (!no_adi)
{
  cout << "DiagLaplaceEqui2D :: FactorADIMatrices()" << endl;
  cout << "# # # # #    d i a c t i v a t e d    # # # # #" << endl;
}
else
{
 if (diagx!=0) delete [] diagx;
 diagx  = new double [nfx+1];
 if (offx!=0) delete [] offx;
 offx   = new double [nfx];

//	rho_x*I = sqrt(2)*Laplace_x + c/2 *I
//old rho = sqrt(8.0)*nfy*nfy;
				//	tridiagonal symmetric matrix
 diagx[0] = diagx[nfx] = 1.0;	// Dirichlet
 for (i=1; i<nfx; i++)		// diagonal entries
  {				// diag = rho*I + D_x
    diagsum=0.0;
    for (j=1; j<nfy; j++)  diagsum += diag[idx(j,i)];
      diagx[i] = diagsum/(nfy-1) + 2*CSQ2*by;
//old    diagx[i] = rho-2*bx;
  }

 offx[0] = offx[nfx-1] = 0.0;	// Dirichlet
 for (i=1; i<nfx-1; i++)	// off-digonal entries
     offx[i] =  bx;

 FactorSymmTriDiag(nfx+1, diagx, offx);

 if (diagy!=0) delete [] diagy;
 diagy  = new double [nfy+1];
 if (offy!=0) delete [] offy;
 offy   = new double [nfy];

//	rho_y*I = sqrt(2)*Laplace_y + c/2 *I
//old rho = sqrt(8.0)*nfx*nfx;
				//	tridiagonal symmetric matrix
 diagy[0] = diagy[nfy] = 1.0;			// Dirichlet
 for (i=1; i<nfy; i++)
  {
    diagsum=0.0;
    for (j=1; j<nfx; j++)  diagsum += diag[idx(i,j)];
    diagy[i] = diagsum/(nfx-1)+2*CSQ2*bx;		// diagonal
//old    diagy[i] = rho-2*by;				// diagonal
  }

 offy[0] = offy[nfy-1] = 0.0;			// Dirichlet
 for (i=1; i<nfy-1; i++)
     offy[i] =  by;					// off-digonal

 FactorSymmTriDiag(nfy+1, diagy, offy);
}
 return;
}

//##############################################################################
//##############################################################################

//	###################################################################
//			LaplaceNonEqui2D
//	###################################################################
//	Laplace 2D, non-equidistant tensor product grid

//---------------------------------------------------------------------------

LaplaceNonEqui2D ::
LaplaceNonEqui2D(const int nnx, const double x[],
                 const int nny, const double y[], const bool coarse)
 : NonEqui2D(nnx,x,nny,y),
   diag(0), bxr(0), bxl(0), byr(0), byl(0), diaga(0), offa(0)
{
//   diag = diaga = 0;
  InitOperator(coarse);
}

//---------------------------------------------------------------------------

LaplaceNonEqui2D ::
LaplaceNonEqui2D(const int nnx, const double *x, const bool coarse)
 : NonEqui2D(nnx,x),
   diag(0), bxr(0), bxl(0), byr(0), byl(0), diaga(0), offa(0)
{
//   if (x==0)
//    {
//      diag = bxr = bxl = byr = byl = diaga = offa = 0;
//    }
//   else
//    {
//      diag = diaga = offa = 0;
     InitOperator(coarse);
//    }
}
//---------------------------------------------------------------------------
LaplaceNonEqui2D ::
LaplaceNonEqui2D(const int intx, const int inty,
                 const double xh[], const double yh[],
	         const bool coarse)
 : NonEqui2D(intx, inty, xh, yh),
   diag(0), bxr(0), bxl(0), byr(0), byl(0), diaga(0), offa(0)
{
//   diag = diaga = offa = 0;
  InitOperator(coarse);
}
//---------------------------------------------------------------------------

 LaplaceNonEqui2D :: ~LaplaceNonEqui2D()
 {
  DeleteGridVector(diag);
  delete [] bxl; delete [] bxr;
  delete [] byl; delete [] byr;
  if (offa!=0)  delete [] offa;
  if (diaga!=0) delete [] diaga;
 }

//---------------------------------------------------------------------------

LaplaceNonEqui2D :: LaplaceNonEqui2D(const LaplaceNonEqui2D &orig)
: NonEqui2D(orig),
   diag(0), bxr(0), bxl(0), byr(0), byl(0), diaga(0), offa(0)
{
 NewGridVector(diag);
 VDcopy( VectorLength(), diag, 1, orig.diag, 1 );

 bxl = new double [GetIntX()+1];
 bxr = new double [GetIntX()+1];
 VDcopy( GetIntX()+1, bxl, 1, orig.bxl, 1 );
 VDcopy( GetIntX()+1, bxr, 1, orig.bxr, 1 );
 byl = new double [GetIntY()+1];
 byr = new double [GetIntY()+1];
 VDcopy( GetIntY()+1, byl, 1, orig.byl, 1 );
 VDcopy( GetIntY()+1, byr, 1, orig.byr, 1 );

 int n = (GetIntX() > GetIntY()) ? GetIntX() : GetIntY();
 diaga = new double [n+1];
 offa  = new double [n];
}

//---------------------------------------------------------------------------
LaplaceNonEqui2D& LaplaceNonEqui2D :: operator=(const LaplaceNonEqui2D & orig)
{
 if ( this != &orig )
 {
   NonEqui2D::operator=(orig);

  if (diag != 0)
   {
     DeleteGridVector(diag);
     delete [] bxl; delete [] bxr;
     delete [] byl; delete [] byr;
     if (diaga!=0) delete [] diaga;
     if (offa!=0) delete [] offa;
   }

  if (orig.diag != 0)
   {
     NewGridVector(diag);
     VDcopy( VectorLength(), diag, 1, orig.diag, 1 );
//  PrintVector(orig.VectorLength() , orig.diag);

     bxl = new double [GetIntX()+1];
     bxr = new double [GetIntX()+1];
     VDcopy( GetIntX()+1, bxl, 1, orig.bxl, 1 );
     VDcopy( GetIntX()+1, bxr, 1, orig.bxr, 1 );
     byl = new double [GetIntY()+1];
     byr = new double [GetIntY()+1];
     VDcopy( GetIntY()+1, byl, 1, orig.byl, 1 );
     VDcopy( GetIntY(), byr, 1, orig.byr, 1 );

     int n = (GetIntX() > GetIntY()) ? GetIntX() : GetIntY();
     diaga = new double [n+1];
     offa  = new double [n];
   }
 }
 return *this;
}

//---------------------------------------------------------------------------
void LaplaceNonEqui2D :: Refine(LaplaceNonEqui2D & fine) const
{
  LaplaceNonEqui2D tmpgrid;
  NonEqui2D::Refine(tmpgrid);

  tmpgrid.coarsegrid_ = this;
  tmpgrid.diag = 0;
  tmpgrid.InitOperator(false);

  fine = tmpgrid;   // FIXXXXXXXXIT:  total unsauber

//  fine.DeriveMatrix()  must not be executed since this is already a fine grid ;
}

//---------------------------------------------------------------------------
void LaplaceNonEqui2D ::OffDiag1D
	(const int n, const double h[], double bl[], double br[])
{
 int i;

 br[0] = 0.0;						// Dirichlet-RB
// br[0] = -2.0/(h[0]+h[1])/h[1] = -2.0/h[1]/h[1];	// Neumann-RB
 bl[0] = 0.0;
 for  (i=1; i < n; i++)
  {
//				[Grossmann/Roos] ==> non-symmetric matrix
//   br[i] = -2.0/(h[i]+h[i+1])/h[i+1];
//   bl[i] = -2.0/(h[i]+h[i+1])/h[i];
//				Bloedsinn, aber symmetrisch
   br[i] = -1.0/h[i+1]/h[i+1];
   bl[i] = -1.0/h[i]/h[i];
//   cout <<  i << "::: "<< bl[i] << " "<< br[i]<< "          " << h[i] << " " << h[i+1] << endl;
//				Bloedsinn, aber symmetrisch
//   br[i] = -2.0/(h[i+1]+h[i+2])/h[i+1];
//   bl[i] = -2.0/(h[i]+h[i+1])/h[i];
  }
 br[n] = 0.0;
 bl[n] = 0.0;						// Dirichlet-RB
// bl[n] = -2.0/(h[n]+h[n+1])/h[n] = -2.0/h[n]/h[n];	// Neumann-RB
 return;
}
//---------------------------------------------------------------------------

void LaplaceNonEqui2D ::InitOperator(const bool coarse, const double *cc)
{
  int n   = VectorLength();
  int nhx = GetIntX(),  nhy = GetIntY();
  int j,k,ijk;
  double tj;

  if (diag==0)
   {
//      DeleteGridVector(diag);
//      delete [] bxl; delete [] bxr; delete [] byl; delete [] byr;
     NewGridVector(diag);
     bxl = new double [nhx+1];
     bxr = new double [nhx+1];
     byl = new double [nhy+1];
     byr = new double [nhy+1];
   }

  OffDiag1D(nhx, hhx, bxl, bxr);
  OffDiag1D(nhy, hhy, byl, byr);

//  VDcopyC( n, diag, 0.0 );
  for (j=0, ijk=0; j<=nhy; j++)
   {
    tj = byl[j]+byr[j];
    for (k=0; k<=nhx; k++, ijk++)
     {
       diag[ijk] = -bxl[k]-bxr[k] - tj;
     }
   }

  if  (cc !=0 ) 	//			Add Helmholtz  term
  	VDadd( n, diag, cc );

  SetBoundaryConst(diag, 1.0);		// Whole boundary is Dirichlet
//#define DEBUG_MATRIX
#ifdef DEBUG_MATRIX
  cout << endl << "Fine Matrix" << endl;
  PrintVector(VectorLength(),diag);
  PrintVector(nhx+1,bxl);
  PrintVector(nhx+1,bxr);
  PrintVector(nhy+1,byl);
  PrintVector(nhy+1,byr);
#endif

  if (coarse)
    {			//	diag is used in matrix calculation
      DeriveMatrix();
#ifdef DEBUG_MATRIX
      cout << endl << "Original Coarse Matrix" << endl;
      PrintMatrix(n, mat);
      bool bb=CheckMatrixSymmetry(n, mat);
      cout << "############################" << endl;
#endif
#undef DEBUG_MATRIX

      InvertMatrix();
#ifdef DEBUG_MATRIX
      cout << endl << "Inverted Coarse Matrix" << endl;
      PrintMatrix(n, mat);
#endif
    }

  if (diaga==0)
   {
     int nn = (GetIntX() > GetIntY()) ? GetIntX() : GetIntY();
     diaga = new double [nn+1];
     offa  = new double [nn];
   }

  CalcIntWeights(); 	// matrix dependent interpolation weights
  return;
}

//---------------------------------------------------------------------------
//	Derive matrix for the 7 point stencil
//		and Dirichlet boundary conditions

void LaplaceNonEqui2D :: DeriveMatrix()
{
 unsigned int j,k,n1,ijk,id;
 unsigned int nn = VectorLength();
 unsigned long int nn2 = nn*nn;

 n1 = Dim1();

//	Allocate memory for fully occupied matrix if called the first time
//			(Assumes no changes in dimensions!!)
 if (mat==0)   mat = new double[nn2];
 VDcopyC(nn2, mat, 0.0);	// set matrix to Zero
	// Initialize diagonal with 1  <== Dirichlet boundary
 for ( j=0; j<nn; j++ )  mat[j*(nn+1)] = 1.0;


 for ( j=1; j<ny; j++ )
  {
   for ( k=1; k<nx; k++ )
    {
      ijk = idx(j, k);
      id  = ijk*nn + ijk;
      mat[id] = diag[ijk];
      if ( k != 1    ) mat[id-1 ] = bxl[k];
      if ( k != nx-1 ) mat[id+1 ] = bxr[k];
      if ( j != 1    ) mat[id-n1] = byl[j];
      if ( j != ny-1 ) mat[id+n1] = byr[j];
    }
  }
#ifdef DEBUG_MATRIX
  cout << endl << "Coarse Matrix" << endl;
  PrintMatrix(nn,mat);
#endif
 return;
}

//---------------------------------------------------------------------------
//		Laplace; Dirichlet B.C.

void LaplaceNonEqui2D ::
ProjectOperator(const bool coarse, const LaplaceNonEqui2D & fine)
{
 double tj;
//  int n   = VectorLength();
 int nhx = GetIntX(),  nhy = GetIntY();
 int j,k,ijk;

 ProjectOffDiag1D(nhx+1,bxl,bxr,fine.bxl,fine.bxr,
 			fine.weight_xl,fine.weight_xr);
 ProjectOffDiag1D(nhy+1,byl,byr,fine.byl,fine.byr,
 			fine.weight_yl,fine.weight_yr);

 for (j=0, ijk=0; j<=nhy; j++)
   {
    tj = byl[j]+byr[j];
    for (k=0; k<=nhx; k++, ijk++)
     {
       diag[ijk] = -bxl[k]-bxr[k] - tj;
     }
   }
 SetBoundaryConst(diag, 1.0);		// Whole boundary is Dirichlet
//#define DEBUG_MATRIX
#ifdef DEBUG_MATRIX
  cout << endl << "Fine Matrix" << endl;
  PrintVector(VectorLength(),diag);
  PrintVector(nhx+1,bxl);
  PrintVector(nhx+1,bxr);
  PrintVector(nhy+1,byl);
  PrintVector(nhy+1,byr);
#endif
 if (coarse)
    {			//	diag is used in matrix calculation
      DeriveMatrix();
#ifdef DEBUG_MATRIX
      cout << endl << "Original Coarse Matrix" << endl;
      PrintMatrix(n, mat);
      bool bb=CheckMatrixSymmetry(n, mat);
      cout << "############################" << endl;
#endif
#undef DEBUG_MATRIX

      InvertMatrix();
#ifdef DEBUG_MATRIX
      cout << endl << "Inverted Coarse Matrix" << endl;
      PrintMatrix(n, mat);
#endif
    }

 return;
}

//---------------------------------------------------------------------------
//		Laplace; Dirichlet B.C.

void LaplaceNonEqui2D :: ProjectOffDiag1D
                       (const int nnc, double cbl[], double cbr[],
  			const double fbl[], const double fbr[],
			const double wl[], const double wr[] )
{
 int i,ii;

 ii = 1;		// indicates always a fine node
 for (i=1; i<nnc; i++)
  {
   cbl[i] = wl[ii]*fbr[ii] + wr[ii]*fbl[ii]
            -(fbr[ii]+fbl[ii])*wl[ii]*wr[ii];
   cbr[i-1] = cbl[i];
   ii += 2;
  }

 return;
}

//---------------------------------------------------------------------------
// 		Dirichlet

void LaplaceNonEqui2D :: CalcIntWeights()
{
 CalcIntWeights1D(GetIntX()+1, bxl, bxr, weight_xl, weight_xr);
 CalcIntWeights1D(GetIntY()+1, byl, byr, weight_yl, weight_yr);
// cout << "HH" << endl;
 return;
}

//---------------------------------------------------------------------------
// 		Dirichlet

void LaplaceNonEqui2D :: CalcIntWeights1D
                       (const int nnodes, const double bl[], const double br[],
			double wl[], double wr[] )
{
 int i;

 for (i=1; i<nnodes; i++)
  {
   wl[i] = bl[i]/(bl[i]+br[i]);
   wr[i] = 1.0-wl[i];
  }

 return;
}


//---------------------------------------------------------------------------

void LaplaceNonEqui2D :: DirectSolve(double *u, const double *f) const
{
 if ( nx != 2 ||  ny != 2)
  {
   NonEqui2D::DirectSolve(u,f);
  }
 else
  {
   u[4] = f[4] / diag[4];
  }
}

//---------------------------------------------------------------------------
//
//
//#define NEUER_VERSUCH
#ifdef NEUER_VERSUCH
void LaplaceNonEqui2D :: DefRes(const double *fh, const double *uh, double *DH) const
{
  const double  CW = 1./4.;			//  normal weight

  double 	t,t1,t2zx,t2zy,t4z;
  unsigned int	nf1,nc1,ncx,ncy,j,k,jj,kk,ijk,IJK;
#ifdef REMOVE_MEM_BUG
          bool  bj;
#endif
  const LaplaceNonEqui2D   &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  nf1 = Dim1();
  nc1 = coarsegrid.Dim1();

  VDcopyC(coarsegrid.VectorLength(), DH, 0.0);

  j = 2;
  for (jj=1; jj <= ncy; jj++)
   {
#ifdef REMOVE_MEM_BUG
     bj = ( j != ny);
#endif
     k = 2;
     IJK = IDX(jj, 1);
     for (kk=1; kk < ncx; kk++, IJK++)
     {
 //			Compiler besser untersttzen!!
 //
       ijk = idx(j-1,k-1);
       t   = byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	     bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
       t4z = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;

       ijk = idx(j-1,k);
       t   = byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	     bxl[k  ]*uh[ijk-1]   + bxr[k  ]*uh[ijk+1];
       t2zx = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;

       ijk = idx(j,k-1);
#ifdef REMOVE_MEM_BUG
       t   = byl[j  ]*uh[ijk-nf1] + byr[j  ]* ( bj ? uh[ijk+nf1] : 0.0 ) +
	     bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
#else
       t   = byl[j  ]*uh[ijk-nf1] + byr[j  ]*uh[ijk+nf1] +
	     bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
#endif
       t2zy = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
 //h	if (j==ny)  t2zy=0.0;		// inefficient

       ijk = idx(j,k);
#ifdef REMOVE_MEM_BUG
       t   = byl[j]*uh[ijk-nf1] + byr[j]* ( bj ? uh[ijk+nf1] : 0.0 ) +
	     bxl[k]*uh[ijk-1]   + bxr[k]* uh[ijk+1];
#else
       t   = byl[j]*uh[ijk-nf1] + byr[j]*uh[ijk+nf1] +
	     bxl[k]*uh[ijk-1]   + bxr[k]*uh[ijk+1];
#endif
       t1  = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
 //h	if (j==ny || k==nx)  t1=0.0;		// inefficient

       DH[IJK-nc1-1]     += wxl(k-1)*wyl(j-1)         *t4z;
       DH[IJK-nc1]       += wxr(k-1)*wyl(j-1)         *t4z
			   +         wyl(j-1)         *t2zx;
       DH[IJK-1]         += wxl(k-1)*wyr(j-1)         *t4z
			   +wxl(k-1)                  *t2zy;
       DH[IJK]           += wxr(k-1)*wyr(j-1)         *t4z
			   +         wyr(j-1)         *t2zx
			   +wxr(k-1)                  *t2zy
			   +t1;
       k += 2;
     }
//		k == nx
     {
     ijk = idx(j-1,k-1);
     t   = byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	   bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
     t4z = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;

     ijk = idx(j-1,k);
     t   = byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	   bxl[k  ]*uh[ijk-1];
     t2zx = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;

     ijk = idx(j,k-1);
#ifdef REMOVE_MEM_BUG
     t   = byl[j  ]*uh[ijk-nf1] + byr[j  ]* ( bj ? uh[ijk+nf1] : 0.0 ) +
           bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
#else
     t   = byl[j  ]*uh[ijk-nf1] + byr[j  ]*uh[ijk+nf1] +
	   bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
#endif
     t2zy = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;

     ijk = idx(j,k);
#ifdef REMOVE_MEM_BUG
     t   = byl[j]*uh[ijk-nf1] + byr[j]* ( bj ? uh[ijk+nf1] : 0.0 ) +
	   bxl[k]*uh[ijk-1];
#else
     t   = byl[j]*uh[ijk-nf1] + byr[j]*uh[ijk+nf1] +
	   bxl[k]*uh[ijk-1];
#endif
     t1  = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;

     DH[IJK-nc1-1]     += wxl(k-1)*wyl(j-1)         *t4z;
     DH[IJK-nc1]       += wxr(k-1)*wyl(j-1)         *t4z
	                 +         wyl(j-1)         *t2zx;
     DH[IJK-1]         += wxl(k-1)*wyr(j-1)         *t4z
	        	 +wxl(k-1)                  *t2zy;
     DH[IJK]           += wxr(k-1)*wyr(j-1)         *t4z
			 +         wyr(j-1)         *t2zx
			 +wxr(k-1)                  *t2zy
			 +t1;
     }
     j += 2;
   }

//  CheckBoundaryZero(uh);			// OK
//  coarsegrid.CheckBoundaryZero(DH);	// boundary is not zero !!
//					Set boundary to Zero !!
  coarsegrid.SetBoundaryZero(DH);
}
#else
//---------------------------------------------------------------------------
//
//

void LaplaceNonEqui2D :: DefRes(const double *fh, const double *uh, double *DH) const
{
  const double  CW = 1./4.;			//  normal weight

  double 	t,t1,t2zx,t2zy,t4z;
  unsigned int	nf1,nc1,ncx,ncy,j,k,jj,kk,ijk,IJK;
#ifdef REMOVE_MEM_BUG
          bool  bj,bk;
#endif
  const LaplaceNonEqui2D   &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  nf1 = Dim1();
  nc1 = coarsegrid.Dim1();

  VDcopyC(coarsegrid.VectorLength(), DH, 0.0);

  j = 2;
  for (jj=1; jj <= ncy; jj++)
   {
#ifdef REMOVE_MEM_BUG
     bj = ( j != ny);
#endif
     k = 2;
     IJK = IDX(jj, 1);
     for (kk=1; kk <= ncx; kk++, IJK++)
     {
 //			Compiler besser untersttzen!!
 //
#ifdef REMOVE_MEM_BUG
       bk = ( k != nx);
#endif
       ijk = idx(j-1,k-1);
       t   = byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	     bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
       t4z = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;

       ijk = idx(j-1,k);
#ifdef REMOVE_MEM_BUG
       t   = byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	     bxl[k  ]*uh[ijk-1]   + bxr[k  ]* ( bk ? uh[ijk+1] : 0.0 ) ;
#else
       t   = byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	     bxl[k  ]*uh[ijk-1]   + bxr[k  ]*uh[ijk+1];
#endif
       t2zx = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
 //h	if (k==nx)  t2zx=0.0;		// inefficient

       ijk = idx(j,k-1);
#ifdef REMOVE_MEM_BUG
       t   = byl[j  ]*uh[ijk-nf1] + byr[j  ]* ( bj ? uh[ijk+nf1] : 0.0 ) +
	     bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
#else
       t   = byl[j  ]*uh[ijk-nf1] + byr[j  ]*uh[ijk+nf1] +
	     bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
#endif
       t2zy = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
 //h	if (j==ny)  t2zy=0.0;		// inefficient

       ijk = idx(j,k);
#ifdef REMOVE_MEM_BUG
       t   = byl[j]*uh[ijk-nf1] + byr[j]* ( bj ? uh[ijk+nf1] : 0.0 ) +
	     bxl[k]*uh[ijk-1]   + bxr[k]* ( bk ? uh[ijk+1] : 0.0 );
#else
       t   = byl[j]*uh[ijk-nf1] + byr[j]*uh[ijk+nf1] +
	     bxl[k]*uh[ijk-1]   + bxr[k]*uh[ijk+1];
#endif
       t1  = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
 //h	if (j==ny || k==nx)  t1=0.0;		// inefficient

       DH[IJK-nc1-1]     += wxl(k-1)*wyl(j-1)         *t4z;
       DH[IJK-nc1]       += wxr(k-1)*wyl(j-1)         *t4z
			   +         wyl(j-1)         *t2zx;
       DH[IJK-1]         += wxl(k-1)*wyr(j-1)         *t4z
			   +wxl(k-1)                  *t2zy;
       DH[IJK]           += wxr(k-1)*wyr(j-1)         *t4z
			   +         wyr(j-1)         *t2zx
			   +wxr(k-1)                  *t2zy
			   +t1;
       k += 2;
     }
     j += 2;
   }

//  CheckBoundaryZero(uh);			// OK
//  coarsegrid.CheckBoundaryZero(DH);	// boundary is not zero !!
//					Set boundary to Zero !!
  coarsegrid.SetBoundaryZero(DH);
}
#endif
//---------------------------------------------------------------------------
//		What happens when we program this method in vector style !??!

void LaplaceNonEqui2D :: Mult(double *f, const double *u) const
{
 double 	t,tjl,tjr;
 unsigned int	j,k,n1,jj;

 n1 = Dim1();

 for (jj=0;  jj<=GetIntX();  jj++)  f[jj] = u[jj];
 for (j=1; j < ny; j++)
 {
   tjl = byl[j];  tjr = byr[j];
   jj  = idx(j,1);
   f[jj-1] = u[jj-1];		// Dirichlet
   for (k=1; k < nx; k++, jj++)
   {
     t = tjl   * u[jj-n1] + tjr   * u[jj+n1] +
	 bxl[k]* u[jj- 1] + bxr[k]* u[jj +1];

     f[jj] = diag[jj] * u[jj] + t;
   }
   f[jj] = u[jj];		// Dirichlet
 }
 for (jj=idx(GetIntY(),0);  jj<=idx(GetIntY(),GetIntX());  jj++)  f[jj] = u[jj];
 return;
}

//---------------------------------------------------------------------------
//		What happens when we program this method in vector style !??!

void LaplaceNonEqui2D ::
Defect(const double *f, const double *u, double *def) const
{
 double 	t,tjl,tjr;
 unsigned int	j,k,n1,jj;

 SetBoundaryZero(def);			// homogeneous Dirichlet B.C.

 n1 = Dim1();

 for (j=1; j < ny; j++)
 {
   tjl = byl[j];  tjr = byr[j];
   jj  = idx(j,1);
   for (k=1; k < nx; k++, jj++)
   {
     t = tjl   * u[jj-n1] + tjr   * u[jj+n1] +
	 bxl[k]* u[jj- 1] + bxr[k]* u[jj +1];

     def[jj] = f[jj] - diag[jj] * u[jj] - t;
   }
 }
 return;
}

//---------------------------------------------------------------------------

void LaplaceNonEqui2D :: Jacobi(double *u, const double *f, double *def,
  			const double omega, const int iter) const
{
 int n = VectorLength();

 Defect(f,u,def);
 VDdiv(n,def,diag);
 VDXpaY(n,u,omega,def);
}

//---------------------------------------------------------------------------

void LaplaceNonEqui2D ::
GaussSeidelForward(double *u, const double *f, const int iter) const
{
  double 	t,tjl,tjr;
  unsigned int	j,k,jj,n1;

  n1 = Dim1();

  for (j=1; j < ny; j++)
  {
    tjl = byl[j];  tjr = byr[j];
    jj = idx(j,1);
    for (k=1; k < nx; k++, jj++)
    {
      t = tjl   * u[jj-n1] + tjr   * u[jj+n1] +
	  bxl[k]* u[jj- 1] + bxr[k]* u[jj +1];

      u[jj]= (f[jj] - t) / diag[jj];
    }
  }
#ifdef DEBUG
 CheckBoundaryZero(u);
#endif
}
//---------------------------------------------------------------------------

void LaplaceNonEqui2D ::
GaussSeidelBackward(double *u, const double *f, const int iter) const
{
  double 	t,tjl,tjr;
  unsigned int	j,k,n1,jj;

  n1 = Dim1();

  for (j=ny-1; j > 0; j--)
  {
    tjl = byl[j];  tjr = byr[j];
    jj = idx(j,nx-1);
    for (k=nx-1; k > 0; k--, jj--)
    {
      t = tjl   * u[jj-n1] + tjr   * u[jj+n1] +
	  bxl[k]* u[jj- 1] + bxr[k]* u[jj +1];

      u[jj]= (f[jj] - t) / diag[jj];
    }
  }
}

//---------------------------------------------------------------------------

double LaplaceNonEqui2D :: DefectNorm(const double *u, const double *f) const
{
  double 	t,s,r,tjl,tjr;
  unsigned int	j,k, n1,jj;

  n1 = Dim1();

  r = 0.0;
  for (j=1; j < ny; j++)
  {
    tjl = byl[j];  tjr = byr[j];
    jj = idx(j,1);
    for (k=1; k < nx; k++, jj++)
    {
      t = tjl   * u[jj-n1] + tjr   * u[jj+n1] +
	  bxl[k]* u[jj- 1] + bxr[k]* u[jj +1];

      s = f[jj] - diag[jj] * u[jj] - t;
      r += s*s;
    }
  }

  return(sqrt(r));
 }

//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., Laplace + c
//	rho*I = sqrt(2)*Laplace_y + c/2 *I
//

void LaplaceNonEqui2D :: ADIx(double u[], const double f[]) const
{
  const double  CSQ2 = 1.0-sqrt(2.0);
  double 	tm;
  unsigned int	i,j,jj,n1,nfx,nfy;

  n1  = Dim1();
  nfx = GetIntX();  nfy = GetIntY();

  for (i=0; i <= nfx; i++)   u[i] = 0.0;
  for (i=0; i <= nfx; i++)  tt[i] = 0.0;
  for (j=1; j < nfy; j++)
  {
    jj  = idx(j,0);
//old    rho = sqrt(2.0)*(-byl[j]-byr[j]);

    diaga[0] = diaga[nfx] = 1.0;	// Dirichlet
    for (i=1; i<nfx; i++)		// diag = rho*I + D_x
       diaga[i] = diag[jj+i] + CSQ2*(byl[j]+byr[j]);
//old       diaga[i] = rho-bxl[i]-bxr[i];

    offa[0] = offa[nfx-1] = 0.0;	// Dirichlet
    for (i=1; i<nfx-1; i++)
	offa[i] =  bxl[i+1];		// off-digonal

    FactorSymmTriDiag(nfx+1, diaga, offa);

    tm = u[jj];			//	right hand side
    u[jj] = 0.0;			// Dirichlet
    for (i=1; i < nfx; i++)
      {					// f + (rho-D_y)tm ...
	tm   = u[jj+i];
	u[jj+i] = f[jj+i] +
		CSQ2*(byl[j]+byr[j])*tm - byr[j]*u[jj+n1+i]-byl[j]*tt[i];
//old		(rho+byl[j]+byr[j])*tm - byr[j]*u[jj+n1+i]-byl[j]*tt[i];
	tt[i] = tm;
      }
    u[jj+nfx] = 0.0;

    SubstFactorSymmTriDiag(nfx+1, diaga, offa, u+jj);
  }
  for (i=idx(nfy,0); i <= idx(nfy,nfx); i++)  u[i] = 0.0;

  return;
}
//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., Laplace + c
//	rho*I = sqrt(2)*Laplace_x + c/2 *I

void LaplaceNonEqui2D :: ADIy(double u[], const double f[]) const
{
  const double  CSQ2 = 1.0-sqrt(2.0);
  double 	tb,tm;
  unsigned int	i,j,jj,n1,nfx,nfy;

  n1  = Dim1();
  nfx = GetIntX();  nfy = GetIntY();

  for (i=0; i <= nfx; i++)  u[i] = 0.0;
  for (j=1; j < nfy; j++)
  {
    jj = idx(j,0);		//	right hand side
    tm = u[jj];
    u[jj] = 0.0;			// Dirichlet
    for (i=1; i < nfx; i++)
     {					// f + (rho-D_x)tm ...
//old       rho = sqrt(2.0)*(-bxl[i]-bxr[i]);
       tb = tm;
       tm = u[jj+i];
       u[jj+i] = f[jj+i] +
		CSQ2*(bxl[i]+bxr[i])*tm - bxr[i]*u[jj+i+1]-bxl[i]*tb;
//old                (rho+bxl[i]+bxr[i])*tm - bxr[i]*u[jj+i+1]-bxl[i]*tb;
     }
    u[jj+nfx] = 0.0;			// Dirichlet
  }
  for (i=idx(nfy,0); i <= idx(nfy,nfx); i++)  u[i] = 0.0;

  for (i=1; i < nfx; i++)
  {
//old    rho = sqrt(2.0)*(-bxl[i]-bxr[i]);

    diaga[0] = diaga[nfy] = 1.0;	// Dirichlet
    for (j=1; j<nfy; j++)		// diag = rho*I + D_y
       diaga[j] = diag[idx(j,i)] + CSQ2*(bxl[i]+bxr[i]);
//old       diaga[j] = rho-byl[j]-byr[j];

    offa[0] = offa[nfy-1] = 0.0;	// Dirichlet
    for (j=1; j<nfy-1; j++)
	offa[j] =  byl[j+1];		// off-digonal

    FactorSymmTriDiag(nfy+1, diaga, offa);
    SubstFactorSymmTriDiagT(nfy+1, diaga, offa, u+i, n1);
  }

  return;
}
//---------------------------------------------------------------------------

void LaplaceNonEqui2D ::
ADIForward(double *u, const double *f, double *aux, const int iter) const
{
#ifdef DEBUG
  cout << "#########  u_in  ###########" << endl;
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
  cout << "#########  f  ###########" << endl;
  PrintVector(VectorLength(),f);
  CheckBoundaryZero(f);
#endif

  ADIx(u, f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
#endif

  ADIy(u, f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
#endif

  return;
}
//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., only Laplace

void LaplaceNonEqui2D ::
ADIBackward(double *u, const double *f, double *aux, const int iter) const
{
#ifdef DEBUG
  cout << "#########  u_in  ###########" << endl;
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
  cout << "#########  f  ###########" << endl;
  PrintVector(VectorLength(),f);
  CheckBoundaryZero(f);
#endif

  ADIy(u, f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
#endif

  ADIx(u, f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
#endif

  return;
}

//---------------------------------------------------------------------------


//	###################################################################
//			DiagLaplaceEqui2DNeumann
//	###################################################################
//	Laplace 2D, equidistant tensor product grid, homogeneoeus Neumann BC

int DiagLaplaceEqui2DNeumann :: count_defect=0;


DiagLaplaceEqui2DNeumann ::
DiagLaplaceEqui2DNeumann(const int mx, const int my, const bool coarse)
 : DiagLaplaceEqui2D(mx,my,coarse)
{
  DeleteGridVector(diag);
  InitOperator(coarse);
}

//---------------------------------------------------------------------------

DiagLaplaceEqui2DNeumann :: DiagLaplaceEqui2DNeumann(const int mx, const bool coarse)
 : DiagLaplaceEqui2D(mx,mx,coarse)
{
  DeleteGridVector(diag);
  InitOperator(coarse);
}

//---------------------------------------------------------------------------

DiagLaplaceEqui2DNeumann :: ~DiagLaplaceEqui2DNeumann()
{
 ;
}
//---------------------------------------------------------------------------

DiagLaplaceEqui2DNeumann :: DiagLaplaceEqui2DNeumann(const DiagLaplaceEqui2DNeumann &orig)
: DiagLaplaceEqui2D(orig)
{
 ;
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui2DNeumann :: Refine(DiagLaplaceEqui2DNeumann & fine) const
{
//						not a coarse grid
  fine = DiagLaplaceEqui2DNeumann(2*nx,2*ny,false);
  fine.coarsegrid_ = this;
//  fine.DeriveMatrix()  must not be executed since this is already a fine grid ;
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui2DNeumann :: Mult(double *f, const double *u) const
{
 double cx,cy,t;
 unsigned int j,k,n1,ijk;

 n1 = nx+1;

   for ( j=0; j<=ny; j++ )
    {
     for ( k=0; k<=nx; k++ )
      {
	cx = bx; cy = by;
	if ( j==0 || j == ny )  { cx = bx/2;}
	if ( k==0 || k == nx )  { cy = by/2;}

	ijk = idx(j,k);
	t = diag[ijk]*u[ijk];
	if ( k != 0  ) t += cx*u[ijk-1];
	if ( k != nx ) t += cx*u[ijk+1];
	if ( j != 0  ) t += cy*u[ijk-n1];
 	if ( j != ny ) t += cy*u[ijk+n1];
	f[ijk] = t;
//	if (j+k==0) cout << diag[ijk] << "  " << cx << "  " << cy << endl;
//	if (i+j+k==0) cout << u[ijk] << "  " << u[ijk+1] << "  " << u[ijk+n1] << "  " << u[ijk+n2] << endl;
     }
    }
}
//---------------------------------------------------------------------------

//void DiagLaplaceEqui2DNeumann :: Mult(double *f, const double *u) const
void DiagLaplaceEqui2DNeumann :: Mult_long(double *f, const double *u) const
{
 double 	t;
 unsigned int	j,k,jj, n1 = Dim1();

//	#	#	#	#	#	#	#	#	#
//	Inner plains

//   				First row
   jj=idx(0,0);
   f[jj] = diag[jj]*u[jj] + 	by*u[jj+n1]/2. +
	   			bx*u[jj +1]/2.;
   for (jj=idx(0,1);  jj<idx(0,nx);  jj++)
   {
     f[jj] = diag[jj]*u[jj] +   by*u[jj+n1] +
	   			bx*(u[jj- 1] + u[jj +1])/2.;
   }
   jj = idx(0,nx);
   f[jj] = diag[jj]*u[jj] + 	by*u[jj+n1]/2. +
	   			bx*u[jj -1]/2.;

//				Inner rows
   for (j=1; j < ny; j++)
   {
     jj = idx(j,0);
     f[jj] = diag[jj]*u[jj] + 	by*(u[jj-n1] + u[jj+n1])/2. +
	   			bx*u[jj +1];
     jj++;
     for (k=1; k < nx; k++, jj++)
     {
       t = by*(u[jj-n1] + u[jj+n1]) +
	   bx*(u[jj- 1] + u[jj +1]);

       f[jj] = diag[jj] * u[jj] + t;
     }
     f[jj] = diag[jj]*u[jj] + 	by*(u[jj-n1] + u[jj+n1])/2. +
	   			bx*u[jj -1];

//				Last row

   jj=idx(ny,0);
   f[jj] = diag[jj]*u[jj] + 	by*u[jj-n1]/2. +
	   			bx*u[jj +1]/2.;
   for (jj=idx(ny,1);  jj<idx(ny,nx);  jj++)
   {
     f[jj] = diag[jj]*u[jj] + 	by*u[jj-n1] +
	   			bx*(u[jj- 1] + u[jj +1])/2.;
   }
   jj = idx(ny,nx);
   f[jj] = diag[jj]*u[jj] + 	by*u[jj-n1]/2. +
	   			bx*u[jj -1]/2.;
 }

 return;
}

//---------------------------------------------------------------------------
//	Derive matrix diagonal for the 7 point stencil
//		and Neumann boundary conditions

void DiagLaplaceEqui2DNeumann :: DeriveDiag(const double *cc)
{
 double cx,cy;
 unsigned int j,k,ijk;
 unsigned int n=VectorLength();

 if (diag==0)   NewGridVector(diag);
// VDcopyC( n, diag, -(bx+by+bz) );

 ijk = 0;
   for ( j=0; j<=ny; j++ )
    {
     for ( k=0; k<=nx; k++, ijk++ )
      {
	cx = bx; cy = by;
	if ( j==0 || j == ny )  { cx = bx/2.; }
	if ( k==0 || k == nx )  { cy = by/2.; }
	diag[ijk] = -(cx+cy);
	if ( k != 0 && k != nx ) diag[ijk] -= cx;
	if ( j != 0 && j != ny ) diag[ijk] -= cy;
     }
    }

 if  (cc !=0 ) 			//	Add Helmholtz  term
       VDadd( n, diag, cc );

#ifdef DEBUG_MATRIX
  cout << endl << "Diagonal in 2DNeumann" << endl;
  PrintVector(n,diag);
#endif
}

//---------------------------------------------------------------------------
//	Derive coarse matrix for the 5 point stencil
//		and Neumann boundary conditions

void DiagLaplaceEqui2DNeumann :: DeriveMatrix()
{
 double cx,cy;
 unsigned int ijk,id, n1= Dim1();
 unsigned int nn = VectorLength();
 unsigned long int nn2 = nn*nn;

//	Allocate memory for fully occupied matrix if called the first time
//			(Assumes no changes in dimensions!!)
 if (mat==0)   mat = new double[nn2];
 VDcopyC(nn2, mat, 0.0);	// set matrix to Zero

   for (unsigned int j=0; j<=ny; ++j )
    {
     for (unsigned int k=0; k<=nx; ++k )
      {
	cx = bx; cy = by;
	if ( j==0 || j == ny )  { cx = bx/2.; }
	if ( k==0 || k == nx )  { cy = by/2.; }

	ijk = idx(j, k);
	id  = ijk*nn + ijk;
	mat[id] = diag[ijk];
	if ( k != 0  ) mat[id-1 ] = cx;
	if ( k != nx ) mat[id+1 ] = cx;
	if ( j != 0  ) mat[id-n1] = cy;
 	if ( j != ny ) mat[id+n1] = cy;
     }
    }
#ifdef DEBUG_MATRIX
  cout << endl << "Coarse Matrix in 2DNeumann" << endl;
  PrintMatrix(nn,mat);
#endif
}


//---------------------------------------------------------------------------

void DiagLaplaceEqui2DNeumann :: Defect(const double *f, const double *u,
					double *def) const
{
 if (def==0)
 {
   cout << endl << "DiagLaplaceEqui2DNeumann :: Defect    def-vector is not allocated!!" << endl;
   exit(-1);
 }
 else
 {
//DB   cout << "Defect" << endl;
   Mult(def,u);
/*//DB
   ++count_defect;
   char name[] = "def_xxx.dat";
   name[4] = 48+count_defect/100;
   name[5] = 48+(count_defect%100)/10;
   name[6] = 48+(count_defect%10);
   if (VectorLength()>9) VectorToFile(def,name);
//BD */
   VDaXpY(VectorLength(),def,-1.0,f);
 }
}
//---------------------------------------------------------------------------

void DiagLaplaceEqui2DNeumann :: DirectSolve(double *u, const double *f) const
{
 Equi2D::DirectSolve(u,f);
}

//---------------------------------------------------------------------------

double DiagLaplaceEqui2DNeumann :: DefectNorm(const double *u, const double *f) const
{
  double *aux,s;
//DB   cout << "Defectnorm" << endl;
  NewGridVector(aux);
  Defect(f,u,aux);
  s = VDnorm(VectorLength(),aux);
  DeleteGridVector(aux);
  return sqrt(s);
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui2DNeumann :: DefRes(const double *fh, const double *uh,
					double *DH) const
{
  double *aux;
  NewGridVector(aux);
  Defect(fh,uh,aux);
  Restrict(aux, DH);
  //DB
/*   ++count_defect;
   char name[] = "def_xxx.dat";
   name[4] = 48+count_defect/100;
   name[5] = 48+(count_defect%100)/10;
   name[6] = 48+(count_defect%10);
   if (VectorLength()>9) VectorToFile(aux,name);*/
//BD
  DeleteGridVector(aux);
}

//---------------------------------------------------------------------------

void DiagLaplaceEqui2DNeumann :: IntAdd(double *uh, const double *WH) const
{
  double 	t,t2,t4;
  unsigned int	nf1,ncx,ncy,j,k,jj,kk,ijk,IJK;
  const Equi2D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  nf1 = Dim1();

    j = 0;
    for (jj=0; jj <= ncy; jj++)
    {
      k = 0;
      ijk = idx(j , k);
      IJK = IDX(jj, 0);
      for (kk=0; kk <= ncx; kk++, IJK++)
      {
	t  = WH[IJK];
	t2 = t/2.;
	t4 = t/4.;

	uh[ijk]	+= t;

	if (j !=  0 ) uh[ijk-nf1]	+= t2;
	if (j != ny ) uh[ijk+nf1]	+= t2;
	if (k !=  0 ) uh[ijk-1]		+= t2;
	if (k != nx ) uh[ijk+1]		+= t2;

	if (j !=  0 && k !=  0 ) uh[ijk-nf1-1]		+= t4;
	if (j !=  0 && k != nx ) uh[ijk-nf1+1]		+= t4;
	if (j != ny && k !=  0 ) uh[ijk+nf1-1]		+= t4;
	if (j != ny && k != nx ) uh[ijk+nf1+1]		+= t4;

        k += 2;  ijk += 2;
      }
      j += 2;
    }
}

//---------------------------------------------------------------------------
void DiagLaplaceEqui2DNeumann :: Inject(const double *uh, double *UH) const
{

  unsigned int	ncx,ncy,j,jj,kk,ijk,IJK;
  const Grid2D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();

    j = 0;
    for (jj=0; jj <= ncy; jj++)
    {
      ijk = idx(j , 0);
      IJK = IDX(jj, 0);
      for (kk=0; kk <= ncx; kk++, IJK++)
      {
	UH[IJK] = uh[ijk];
	ijk += 2;
      }
      j += 2;
    }
}
//---------------------------------------------------------------------------
void DiagLaplaceEqui2DNeumann :: Restrict(const double *uh, double *UH) const
{
//#define INJECT
#ifdef INJECT
 Inject(uh,UH);
#else

  double  CW;				//  normal weight
  double 	t,t2,t4,c2,c4;
  unsigned int	nf1,ncx,ncy,j,k,jj,kk,ijk,IJK;
  const Equi2D  &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  nf1 = Dim1();

    j = 0;
    for (jj=0; jj <= ncy; jj++)
    {
      k = 0;
      ijk = idx(j , k);
      IJK = IDX(jj, 0);
      for (kk=0; kk <= ncx; kk++, IJK++)
      {
        t2 = t4 = 0.0;
        c2 = c4 = 0.0;

	t  = uh[ijk];

        if (j !=  0) {t2 += uh[ijk-nf1]; c2++;}
        if (j != ny) {t2 += uh[ijk+nf1]; c2++;}
        if (k !=  0) {t2 += uh[ijk-  1]; c2++;}
        if (k != nx) {t2 += uh[ijk+  1]; c2++;}

	t2 = t2/2.0;

	if (k !=  0 && j !=  0) {t4 += uh[ijk-nf1-1]; c4++;}
	if (k !=  0 && j != ny) {t4 += uh[ijk+nf1-1]; c4++;}
	if (k != nx && j !=  0) {t4 += uh[ijk-nf1+1]; c4++;}
	if (k != nx && j != ny) {t4 += uh[ijk+nf1+1]; c4++;}

	t4 = t4/4.0;

	CW = 1.0/(1.0 + c2/2. + c4/4.0);
//	cout << CW << endl;
	UH[IJK] = (t+t2+t4) * CW ;

	k += 2;  ijk += 2;
      }
      j += 2;
    }

#endif
}





