#include <iostream>
#include <cmath>
#include <cassert>
using namespace std;
#include "vector.hpp"
#include "grid3d.hpp"
#include "laplace_vary_3d.hpp"

//	###################################################################
//			LaplaceVaryNonEqui3D
//	###################################################################
//	Laplace 3D, non-equidistant tensor product grid, variable coeff.,
//    homog. Dirichlet BC

//---------------------------------------------------------------------------

LaplaceVaryNonEqui3D ::
LaplaceVaryNonEqui3D(const int nnx, const double x[],
                 const int nny, const double y[],
	         const int nnz, const double z[], const bool coarse)
 : NonEqui3D(nnx,x,nny,y,nnz,z),
   diag(0), ak(0), bxr(0), bxl(0), byr(0), byl(0), bzr(0), bzl(0)
{
  InitOperator(coarse);
}

//---------------------------------------------------------------------------

LaplaceVaryNonEqui3D ::
LaplaceVaryNonEqui3D(const int nnx, const double *x, const bool coarse)
 : NonEqui3D(nnx,x),
   diag(0), ak(0), bxr(0), bxl(0), byr(0), byl(0), bzr(0), bzl(0)
{
     InitOperator(coarse);
}
//---------------------------------------------------------------------------
LaplaceVaryNonEqui3D ::
LaplaceVaryNonEqui3D(const int intx, const int inty, const int intz,
                 const double xh[], const double yh[], const double zh[],
	         const bool coarse)
 : NonEqui3D(intx, inty, intz, xh, yh, zh),
   diag(0), ak(0), bxr(0), bxl(0), byr(0), byl(0), bzr(0), bzl(0)
{
  InitOperator(coarse);
}

//---------------------------------------------------------------------------
// LaplaceVaryNonEqui3D ::
// LaplaceVaryNonEqui3D(const NonEqui3D &orig)
//  : NonEqui3D(orig),
//    diag(0), ak(0), bxr(0), bxl(0), byr(0), byl(0), bzr(0), bzl(0)
// {
//   InitOperator(false);
// }

//---------------------------------------------------------------------------

 LaplaceVaryNonEqui3D :: ~LaplaceVaryNonEqui3D()
 {
  DeleteGridVector(diag);
  DeleteGridVector(ak);
  delete [] bxl; delete [] bxr;
  delete [] byl; delete [] byr;
  delete [] bzl; delete [] bzr;
 }

//---------------------------------------------------------------------------

LaplaceVaryNonEqui3D :: LaplaceVaryNonEqui3D(const LaplaceVaryNonEqui3D &orig)
: NonEqui3D(orig),
   diag(0), ak(0), bxr(0), bxl(0), byr(0), byl(0), bzr(0), bzl(0)
{
 NewGridVector(diag);
 VDcopy( VectorLength(), diag, 1, orig.diag, 1 );
 NewGridVector(ak);
 VDcopy( VectorLength(), ak, 1, orig.ak, 1 );

 bxl = new double [GetIntX()+1];
 bxr = new double [GetIntX()+1];
 VDcopy( GetIntX()+1, bxl, 1, orig.bxl, 1 );
 VDcopy( GetIntX()+1, bxr, 1, orig.bxr, 1 );
 byl = new double [GetIntY()+1];
 byr = new double [GetIntY()+1];
 VDcopy( GetIntY()+1, byl, 1, orig.byl, 1 );
 VDcopy( GetIntY()+1, byr, 1, orig.byr, 1 );
 bzl = new double [GetIntZ()+1];
 bzr = new double [GetIntZ()+1];
 VDcopy( GetIntZ()+1, bzl, 1, orig.bzl, 1 );
 VDcopy( GetIntZ()+1, bzr, 1, orig.bzr, 1 );
}

//---------------------------------------------------------------------------
LaplaceVaryNonEqui3D& LaplaceVaryNonEqui3D :: operator=(const LaplaceVaryNonEqui3D & orig)
{
 if ( this != &orig )
 {
   NonEqui3D::operator=(orig);

  if (diag != 0)
   {
     DeleteGridVector(diag);
     DeleteGridVector(ak);
     delete [] bxl; delete [] bxr;
     delete [] byl; delete [] byr;
     delete [] bzl; delete [] bzr;
   }

  if (orig.diag != 0)
   {
     NewGridVector(diag);
     VDcopy( VectorLength(), diag, 1, orig.diag, 1 );
     NewGridVector(ak);
     VDcopy( VectorLength(), ak, 1, orig.ak, 1 );

     bxl = new double [GetIntX()+1];
     bxr = new double [GetIntX()+1];
     VDcopy( GetIntX()+1, bxl, 1, orig.bxl, 1 );
     VDcopy( GetIntX()+1, bxr, 1, orig.bxr, 1 );
     byl = new double [GetIntY()+1];
     byr = new double [GetIntY()+1];
     VDcopy( GetIntY()+1, byl, 1, orig.byl, 1 );
     VDcopy( GetIntY(), byr, 1, orig.byr, 1 );
     bzl = new double [GetIntZ()+1];
     bzr = new double [GetIntZ()+1];
     VDcopy( GetIntZ()+1, bzl, 1, orig.bzl, 1 );
     VDcopy( GetIntZ()+1, bzr, 1, orig.bzr, 1 );
   }
 }
 return *this;
}

//---------------------------------------------------------------------------
void LaplaceVaryNonEqui3D :: Refine(LaplaceVaryNonEqui3D & fine) const
{
//   NonEqui3D tmpgrid;
//   NonEqui3D::Refine(tmpgrid);
//
//   fine = LaplaceVaryNonEqui3D(tmpgrid);
//   fine.coarsegrid_ = this;
//   fine.InitOperator(false);

  LaplaceVaryNonEqui3D tmpgrid;
  NonEqui3D::Refine(tmpgrid);

  tmpgrid.coarsegrid_ = this;
  tmpgrid.diag = 0;
  tmpgrid.ak = 0;
  tmpgrid.InitOperator(false);
  fine = tmpgrid;

//  fine.DeriveMatrix()  must not be executed since this is already a fine grid ;
}

//---------------------------------------------------------------------------
void LaplaceVaryNonEqui3D ::OffDiag1D
	(const int n, const double h[], double bl[], double br[])
{
 int i;

 br[0] = 0.0;						// Dirichlet-RB
// br[0] = -2.0/(h[0]+h[1])/h[1] = -2.0/h[1]/h[1];	// Neumann-RB
 bl[0] = 0.0;
 for  (i=1; i < n; i++)
  {
//				[Grossmann/Roos] ==> non-symmetric matrix
/*  br[i] = -1.0/(h[i]+h[i+1])/h[i+1];
  bl[i] = -1.0/(h[i]+h[i+1])/h[i];*/
//				Bloedsinn, aber symmetrisch
//                      !! Faktor 0.5 von Koeff. wird hier inkludiert !!!
   br[i] = -0.5/h[i+1]/h[i+1];
   bl[i] = -0.5/h[i]/h[i];
//    br[i] = -0.5/h[i+1];
//    bl[i] = -0.5/h[i];

//    cout <<  i << "::: "<< bl[i] << " "<< br[i]<< "          " << h[i] << " " << h[i+1] << endl;
//				Bloedsinn, aber symmetrisch
//   br[i] = -2.0/(h[i+1]+h[i+2])/h[i+1];
//   bl[i] = -2.0/(h[i]+h[i+1])/h[i];
  }
 br[n] = 0.0;
 bl[n] = 0.0;						// Dirichlet-RB
// bl[n] = -2.0/(h[n]+h[n+1])/h[n] = -2.0/h[n]/h[n];	// Neumann-RB
 return;
}
//---------------------------------------------------------------------------

void LaplaceVaryNonEqui3D ::InitOperator(const bool coarse, const double *cc)
{
  int n   = VectorLength();
//FIXXIT:  if (n<=1) return;
  int nhx = GetIntX(),  nhy = GetIntY(),  nhz = GetIntZ();
  int i,j,k,ijk;

  if (diag==0)
   {
     NewGridVector(diag);
     NewGridVector(ak);
     bxl = new double [nhx+1];
     bxr = new double [nhx+1];
     byl = new double [nhy+1];
     byr = new double [nhy+1];
     bzl = new double [nhz+1];
     bzr = new double [nhz+1];
   }

  if  (cc !=0 )        // standard case: coeff. is constant one.
    VDcopyC( n, ak, 1.0 );


  OffDiag1D(nhx, hhx, bxl, bxr);
  OffDiag1D(nhy, hhy, byl, byr);
  OffDiag1D(nhz, hhz, bzl, bzr);

//    Only correct for Dirichlet BC
  VDcopyC( n, diag, 1.0 );
  for (i=1; i<nhz; i++)
   {
    for (j=1; j<nhy; j++)
     {
      ijk = idx(i,j,1);
      for (k=1; k<nhx; k++, ijk++)
       {
         diag[ijk] = - (ak[ijk]+ak[ijk-1]     )*bxl[k] - (ak[ijk+1     ]+ak[ijk])*bxr[k]
                     - (ak[ijk]+ak[ijk-Dim1()])*byl[j] - (ak[ijk+Dim1()]+ak[ijk])*byr[j]
                     - (ak[ijk]+ak[ijk-Dim2()])*bzl[i] - (ak[ijk+Dim2()]+ak[ijk])*bzr[i];
       }
     }
   }

//   if  (cc !=0 ) 	//			Add Helmholtz  term
//   	VDadd( n, diag, cc );

//   SetBoundaryConst(diag, 1.0);		// Whole boundary is Dirichlet

  if (coarse)
    {			//	diag is used in matrix calculation
      DeriveMatrix();
      bool bb=CheckMatrixSymmetry(n, mat);
#ifdef DEBUG_MATRIX
      cout << endl << "Original Coarse Matrix" << endl;
      PrintMatrix(n, mat);
      bool bb=CheckMatrixSymmetry(n, mat);
      cout << "############################" << endl;
#endif
      InvertMatrix();
#ifdef DEBUG_MATRIX
      cout << endl << "Inverted Coarse Matrix" << endl;
      PrintMatrix(n, mat);
#endif
    }
}

//---------------------------------------------------------------------------
//	Derive matrix for the 7 point stencil
//		and Dirichlet boundary conditions

void LaplaceVaryNonEqui3D :: DeriveMatrix()
{
 int i,j,k,n1,n2,ijk,id;
 int nn = VectorLength();
 unsigned long int nn2 = nn*nn;

 n1 = nx+1;  n2 = (nx+1)*(ny+1);

//	Allocate memory for fully occupied matrix if called the first time
//			(Assumes no changes in dimensions!!)
 if (mat==0)   mat = new double[nn2];
 VDcopyC(nn2, mat, 0.0);	// set matrix to Zero
	// Initialize diagonal with 1  <== Dirichlet boundary
 for ( i=0; i<nn; i++ )  mat[i*(nn+1)] = 1.0;

 for ( i=1; i<nz; i++ )
  {
   for ( j=1; j<ny; j++ )
    {
     for ( k=1; k<nx; k++ )
      {
      ijk = idx(n2,n1,i,j, k);
	id  = ijk*nn + ijk;
/*	cout << i << "  " << j << "  " << k << " : " << i << " ijk "
	     << id << "  " << diag[ijk] << endl ;*/
	mat[id] = diag[ijk];
	if ( k != 1    ) mat[id-1 ] = (ak[ijk]+ak[ijk-1]     )*bxl[k];
	if ( k != nx-1 ) mat[id+1 ] = (ak[ijk+1     ]+ak[ijk])*bxr[k];
	if ( j != 1    ) mat[id-n1] = (ak[ijk]+ak[ijk-Dim1()])*byl[j];
 	if ( j != ny-1 ) mat[id+n1] = (ak[ijk+Dim1()]+ak[ijk])*byr[j];
	if ( i != 1    ) mat[id-n2] = (ak[ijk]+ak[ijk-Dim2()])*bzl[i];
 	if ( i != nz-1 ) mat[id+n2] = (ak[ijk+Dim2()]+ak[ijk])*bzr[i];
     }
    }
  }
#ifdef DEBUG_MATRIX
  cout << endl << "Coarse Matrix" << endl;
  PrintMatrix(nn,mat);
#endif
}

//---------------------------------------------------------------------------

void LaplaceVaryNonEqui3D :: DirectSolve(double *u, const double *f) const
{
 if ( nx != 2 ||  ny != 2 || nz != 2)
  {
   NonEqui3D::DirectSolve(u,f);
  }
 else
  {
   u[13] = f[13] / diag[13];
  }
}

//---------------------------------------------------------------------------

void LaplaceVaryNonEqui3D :: DefRes(const double *fh, const double *uh, double *DH) const
{
 const int n = VectorLength();
 double *def;

 NewGridVector(def);
 Defect(fh,uh,def);
 Restrict(def,DH);        // ein eigenes Restrict implementieren !!
                          // ein eigenes Interpolate implementieren !!
 DeleteGridVector(def);

//  CheckBoundaryZero(uh);			// OK
//  coarsegrid.CheckBoundaryZero(DH);	// boundary is not zero !!
//					Set boundary to Zero !!
  CoarseGrid().SetBoundaryZero(DH);
}
//---------------------------------------------------------------------------

void LaplaceVaryNonEqui3D ::
DefRes_fast(const double *fh, const double *uh, double *DH) const
{
  const double  CW = 1./8.;			//  normal weight

  double 	t,t1,t2yx,t2zx,t2zy,t4x,t4y,t4z,t8;
  unsigned int	nf1,nf2,nc1,nc2,ncx,ncy,ncz,i,j,k,ii,jj,kk,ijk,IJK;
  const LaplaceVaryNonEqui3D   &coarsegrid = CoarseGrid();

  ncx = coarsegrid.GetIntX();
  ncy = coarsegrid.GetIntY();
  ncz = coarsegrid.GetIntZ();
  nf1 = Dim1();              nf2 = Dim2();
  nc1 = coarsegrid.Dim1();   nc2 = coarsegrid.Dim2();

  VDcopyC(coarsegrid.VectorLength(), DH, 0.0);

  i = 2;
  for (ii=1; ii <= ncz; ii++)
  {
   j = 2;
   for (jj=1; jj <= ncy; jj++)
    {
      k = 2;
      IJK = idx(nc2,nc1,ii,jj, 1);
      for (kk=1; kk <= ncx; kk++, IJK++)
      {
//			Compiler besser untersttzen!!
//
	ijk = idx(nf2,nf1,i-1,j-1,k-1);
        t   = bzl[i-1]*uh[ijk-nf2] + bzr[i-1]*uh[ijk+nf2] +
	        byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	        bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
        t8  = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;

	ijk = idx(nf2,nf1,i-1,j-1,k);
        t   = bzl[i-1]*uh[ijk-nf2] + bzr[i-1]*uh[ijk+nf2] +
	        byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	        bxl[k  ]*uh[ijk-1]   + bxr[k  ]*uh[ijk+1];
        t4x = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (k==nx)  t4x=0.0;		// inefficient

	ijk = idx(nf2,nf1,i-1,j,k-1);
        t   = bzl[i-1]*uh[ijk-nf2] + bzr[i-1]*uh[ijk+nf2] +
	        byl[j  ]*uh[ijk-nf1] + byr[j  ]*uh[ijk+nf1] +
	        bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
        t4y = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (j==ny)  t4y=0.0;		// inefficient

	ijk = idx(nf2,nf1,i,j-1,k-1);
        t   = bzl[i  ]*uh[ijk-nf2] + bzr[i  ]*uh[ijk+nf2] +
	        byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	        bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
        t4z = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (i==nz)  t4z=0.0;		// inefficient

	ijk = idx(nf2,nf1,i-1,j,k);
        t   = bzl[i-1]*uh[ijk-nf2] + bzr[i-1]*uh[ijk+nf2] +
	        byl[j  ]*uh[ijk-nf1] + byr[j  ]*uh[ijk+nf1] +
	        bxl[k  ]*uh[ijk-1]   + bxr[k  ]*uh[ijk+1];
        t2yx = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (k==nx || j==ny)  t2yx=0.0;		// inefficient

	ijk = idx(nf2,nf1,i,j-1,k);
        t   = bzl[i  ]*uh[ijk-nf2] + bzr[i  ]*uh[ijk+nf2] +
	        byl[j-1]*uh[ijk-nf1] + byr[j-1]*uh[ijk+nf1] +
	        bxl[k  ]*uh[ijk-1]   + bxr[k  ]*uh[ijk+1];
        t2zx = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (i==nz || k==nx)  t2zx=0.0;		// inefficient

	ijk = idx(nf2,nf1,i,j,k-1);
        t   = bzl[i  ]*uh[ijk-nf2] + bzr[i  ]*uh[ijk+nf2] +
	        byl[j  ]*uh[ijk-nf1] + byr[j  ]*uh[ijk+nf1] +
	        bxl[k-1]*uh[ijk-1]   + bxr[k-1]*uh[ijk+1];
        t2zy = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (i==nz || j==ny)  t2zy=0.0;		// inefficient

 	ijk = idx(nf2,nf1,i,j,k);
        t   = bzl[i]*uh[ijk-nf2] + bzr[i]*uh[ijk+nf2] +
	        byl[j]*uh[ijk-nf1] + byr[j]*uh[ijk+nf1] +
	        bxl[k]*uh[ijk-1]   + bxr[k]*uh[ijk+1];
        t1  = (fh[ijk] - diag[ijk] * uh[ijk] - t)*CW;
//h	if (i==nz || j==ny || k==nx)  t1=0.0;		// inefficient

        DH[IJK-nc2-nc1-1] += wxl(k-1)*wyl(j-1)*wzl(i-1)*t8;
        DH[IJK-nc2-nc1]   += wxr(k-1)*wyl(j-1)*wzl(i-1)*t8
			          +         wyl(j-1)*wzl(i-1)*t4x;
        DH[IJK-nc2-1]     += wxl(k-1)*wyr(j-1)*wzl(i-1)*t8
			          +wxl(k-1)         *wzl(i-1)*t4y;
        DH[IJK-nc1-1]     += wxl(k-1)*wyl(j-1)*wzr(i-1)*t8
	                      +wxl(k-1)*wyl(j-1)         *t4z;
        DH[IJK-nc2]       += wxr(k-1)*wyr(j-1)*wzl(i-1)*t8
	                      +         wyr(j-1)*wzl(i-1)*t4x
			          +wxr(k-1)         *wzl(i-1)*t4y
			          +                  wzl(i-1)*t2yx;
        DH[IJK-nc1]       += wxr(k-1)*wyl(j-1)*wzr(i-1)*t8
	                      +         wyl(j-1)*wzr(i-1)*t4x
			          +wxr(k-1)*wyl(j-1)         *t4z
			          +         wyl(j-1)         *t2zx;
        DH[IJK-1]         += wxl(k-1)*wyr(j-1)*wzr(i-1)*t8
	                      +wxl(k-1)         *wzr(i-1)*t4y
			          +wxl(k-1)*wyr(j-1)         *t4z
			          +wxl(k-1)                  *t2zy;
        DH[IJK]           += wxr(k-1)*wyr(j-1)*wzr(i-1)*t8
	                      +         wyr(j-1)*wzr(i-1)*t4x
	                      +wxr(k-1)         *wzr(i-1)*t4y
			          +wxr(k-1)*wyr(j-1)         *t4z
			          +                  wzr(i-1)*t2yx
			          +         wyr(j-1)         *t2zx
			          +wxr(k-1)                  *t2zy
			          +t1;

	k += 2;
      }
      j += 2;
    }
    i += 2;
  }
//  CheckBoundaryZero(uh);			// OK
//  coarsegrid.CheckBoundaryZero(DH);	// boundary is not zero !!
//					Set boundary to Zero !!
  coarsegrid.SetBoundaryZero(DH);
}

// - (ak[ijk]-ak[ijk-1] )*bxl[k] - (ak[ijk+1 ]-ak[ijk])*bxr[k]
// - (ak[ijk]-ak[ijk-n1])*byl[j] - (ak[ijk+n1]-ak[ijk])*byr[j]
// - (ak[ijk]-ak[ijk-n2])*bzl[i] - (ak[ijk+n2]-ak[ijk])*bzr[i]

//---------------------------------------------------------------------------
//		What happens when we program the method in vector style !??!

void LaplaceVaryNonEqui3D :: Mult(double *f, const double *u) const
{
 double 	      t;
 unsigned int	i,j,k,ij;
 const unsigned int n1 = Dim1(), n2 = Dim2();

 for (ij=idx(0,0,0);  ij<=idx(0,ny,nx);  ij++)  f[ij] = u[ij];
 for (i=1; i < nz; i++)
 {
   for (ij=idx(i,0,0);  ij<=idx(i,0,nx);  ij++)  f[ij] = u[ij];
   for (j=1; j < ny; j++)
   {
     ij  = idx(i,j,1);
     f[ij-1] = u[ij-1];		// Dirichlet
     for (k=1; k < nx; k++, ij++)
     {
       t = (ak[ij]+ak[ij-n2])*bzl[i] * u[ij-n2]
         + (ak[ij+n2]+ak[ij])*bzr[i] * u[ij+n2]
         + (ak[ij]+ak[ij-n1])*byl[j] * u[ij-n1]
         + (ak[ij+n1]+ak[ij])*byr[j] * u[ij+n1]
         + (ak[ij]+ak[ij-1] )*bxl[k] * u[ij- 1]
         + (ak[ij+1 ]+ak[ij])*bxr[k] * u[ij +1];

       f[ij] = diag[ij] * u[ij] + t;
     }
     f[ij] = u[ij];		// Dirichlet
   }
   for (ij=idx(i,ny,0);  ij<=idx(i,ny,nx);  ij++)  f[ij] = u[ij];
 }
 for (ij=idx(nz,0,0);  ij<=idx(nz,ny,nx);  ij++)  f[ij] = u[ij];
}

//---------------------------------------------------------------------------
//		What happens when we program the method in vector style !??!

void LaplaceVaryNonEqui3D ::
Defect(const double *f, const double *u, double *def) const
{
 double 	        t;
 unsigned int       i,j,k;
 const unsigned int n1 = Dim1(), n2 = Dim2();

 SetBoundaryZero(def);			// homogeneous Dirichlet B.C.

 for (i=1; i < nz; i++)
 {
   for (j=1; j < ny; j++)
   {
     unsigned int  ij  = idx(n2,n1,i,j,1);
     for (k=1; k < nx; k++, ij++)
     {
       t = (ak[ij]+ak[ij-n2])*bzl[i] * u[ij-n2]
         + (ak[ij+n2]+ak[ij])*bzr[i] * u[ij+n2]
         + (ak[ij]+ak[ij-n1])*byl[j] * u[ij-n1]
         + (ak[ij+n1]+ak[ij])*byr[j] * u[ij+n1]
         + (ak[ij]+ak[ij-1] )*bxl[k] * u[ij- 1]
         + (ak[ij+1 ]+ak[ij])*bxr[k] * u[ij +1];

       def[ij] = f[ij] - diag[ij] * u[ij] - t;
     }
   }
 }
}

//---------------------------------------------------------------------------

void LaplaceVaryNonEqui3D :: Jacobi(double *u, const double *f, double *def,
  			const double omega, const int iter) const
{
 int n = VectorLength();

 Defect(f,u,def);
 VDdiv(n,def,diag);
 VDXpaY(n,u,omega,def);
}

//---------------------------------------------------------------------------

void LaplaceVaryNonEqui3D ::
GaussSeidelForward(double *u, const double *f, const int iter) const
{
  double 	t,til,tir,tjl,tjr;
  unsigned int	i,j,k,n1,n2;

  n1 = Dim1(); n2 = Dim2();

  for (i=1; i < nz; i++)
  {
    for (j=1; j < ny; j++)
    {
      unsigned int ij = idx(n2,n1,i,j,1);
      for (k=1; k < nx; k++, ij++)
      {
       t = (ak[ij]+ak[ij-n2])*bzl[i] * u[ij-n2]
         + (ak[ij+n2]+ak[ij])*bzr[i] * u[ij+n2]
         + (ak[ij]+ak[ij-n1])*byl[j] * u[ij-n1]
         + (ak[ij+n1]+ak[ij])*byr[j] * u[ij+n1]
         + (ak[ij]+ak[ij-1] )*bxl[k] * u[ij- 1]
         + (ak[ij+1 ]+ak[ij])*bxr[k] * u[ij +1];

        u[ij]= (f[ij] - t) / diag[ij];
      }
    }
  }
#ifdef DEBUG
 CheckBoundaryZero(u);
#endif
}

//---------------------------------------------------------------------------

void LaplaceVaryNonEqui3D ::
GaussSeidelBackward(double *u, const double *f, const int iter) const
{
  double 	      t;
  unsigned int	i,j,k,n1,n2;

  n1 = Dim1(); n2 = Dim2();

  for (i=nz-1; i > 0; i--)
  {
    for (j=ny-1; j > 0; j--)
    {
      unsigned int ij = idx(n2,n1,i,j,nx-1);
      for (k=nx-1; k > 0; k--, ij--)
      {
       t = (ak[ij]+ak[ij-n2])*bzl[i] * u[ij-n2]
         + (ak[ij+n2]+ak[ij])*bzr[i] * u[ij+n2]
         + (ak[ij]+ak[ij-n1])*byl[j] * u[ij-n1]
         + (ak[ij+n1]+ak[ij])*byr[j] * u[ij+n1]
         + (ak[ij]+ak[ij-1] )*bxl[k] * u[ij- 1]
         + (ak[ij+1 ]+ak[ij])*bxr[k] * u[ij +1];

        u[ij]= (f[ij] - t) / diag[ij];
      }
    }
  }
}

//---------------------------------------------------------------------------

double LaplaceVaryNonEqui3D :: DefectNorm(const double *u, const double *f) const
{
  double 	    t,s,r;
  unsigned int  i,j,k,n1,n2;

  n1 = Dim1(); n2 = Dim2();

  r = 0.0;
  for (i=1; i < nz; i++)
  {
    for (j=1; j < ny; j++)
    {
      unsigned int ij = idx(n2,n1,i,j,1);
      for (k=1; k < nx; k++, ij++)
      {
       t = (ak[ij]+ak[ij-n2])*bzl[i] * u[ij-n2]
         + (ak[ij+n2]+ak[ij])*bzr[i] * u[ij+n2]
         + (ak[ij]+ak[ij-n1])*byl[j] * u[ij-n1]
         + (ak[ij+n1]+ak[ij])*byr[j] * u[ij+n1]
         + (ak[ij]+ak[ij-1] )*bxl[k] * u[ij- 1]
         + (ak[ij+1 ]+ak[ij])*bxr[k] * u[ij +1];

        s = f[ij] - diag[ij] * u[ij] - t;
        r += s*s;
      }
    }
  }

//   cout << "-------------------"<<endl;
//   for (i=1; i<nz; i++)
//      cout << i << ":: " << bzl[i] << " " << bzr[i]
//       << "            " << hhz[i] << " " << hhz[i+1]
//       << endl;

  return(sqrt(r));
 }

//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., Laplace + c
//	rho*I = sqrt(2)*Laplace_y + c/2 *I

void LaplaceVaryNonEqui3D :: ADIx(double u[], const double f[]) const
{
  assert(1==0);
  return;
}
//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C.,  Laplace + c
//	rho*I = sqrt(2)*Laplace_x + c/2 *I

void LaplaceVaryNonEqui3D :: ADIy(double u[], const double f[]) const
{

  assert(1==0);
  return;
}

//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., only Laplace
//	Factorization (rho !!) is done in FactorADIMatrices()

void LaplaceVaryNonEqui3D ::
ADIForward(double *u, const double *f, double *aux, const int iter) const
{
#ifdef DEBUG
  cout << "#########  u_in  ###########" << endl;
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
  cout << "#########  f  ###########" << endl;
  PrintVector(VectorLength(),f);
  CheckBoundaryZero(f);
#endif

  ADIx(u,f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
#endif

  ADIy(u,f);

#ifdef DEBUG
  PrintVector(VectorLength(),u);
  CheckBoundaryZero(u);
#endif
  return;
}

//---------------------------------------------------------------------------
//			Hom. Dirichlet B.C., only Laplace
//	Factorization (rho !!) is done in FactorADIMatrices()

void LaplaceVaryNonEqui3D ::
ADIBackward(double *u, const double *f, double *aux, const int iter) const
{
 ADIy(u,f);
 ADIx(u,f);
#ifdef DEBUG
 CheckBoundaryZero(u);
#endif
  return;
}
