#include <iostream>
#include <cmath>
using namespace std;
#include "vector.hpp"
#include "grid.hpp"
#include "mgprocs.hpp"

//####################################################################
//		Should be also adapted to class MultiGrid
//####################################################################
// #define DEBUG
void MGM(const Grid &grid, VectorRec *f_data, CtrlRec *f_ctrl)
{
  VectorRec  *c_data;
  CtrlRec    *c_ctrl;
     int     i;
  double     *u,*f, *c_u, *c_f;
//s
  double     *def;
  const Grid &c_grid =  grid.CoarseGrid();

  u  = f_data->u;
  f  = f_data->f;
  def= f_data->def;

  if (grid.IsCoarseGrid())
   {
//			Coarse grid solver
     grid.DirectSolve(u,f);
   }
  else
   {
     c_data = f_data-1;
     c_ctrl = f_ctrl-1;
     c_u = c_data->u;
     c_f = c_data->f;

//#define NORM
#ifdef NORM
  double f0 = grid.DefectNorm(u, f);
#endif
//				Pre-Smoothing
     for (i=1; i<=f_ctrl->vGamma; i++)
//	 { grid.GaussSeidelForward(u,f); }
	 { grid.Smooth(u,f,def,f_ctrl->vGlatt,i); }

#ifdef NORM
cout << " presmooth : " << grid.DefectNorm(u, f)/f0 << endl;
#endif
//				Defect calculation and restriction//
//s     grid.Defect(f, u, def);
//s     grid.Restrict(def, c_f);
//#define DEBUG
#ifdef DEBUG
cout << "check" << endl;
PrintVector(c_grid.VectorLength(),c_f);
cout << "normal" << endl;
PrintVector(grid.VectorLength(),u);
PrintVector(grid.VectorLength(),f);
PrintVector(grid.VectorLength(),def);
PrintVector(c_grid.VectorLength(),c_f);
#endif
//#undef DEBUG

//
     grid.DefRes(f, u, c_f);
//      char name[]="c_f_x.dat";
//      name[4] =
//      c_grid.VectorToFile(c_f,"c_f.dat");

#ifdef DEBUG
cout << "combined" << endl;
PrintVector(c_grid.VectorLength(),c_f);
#endif

//				Initialize coarse grid correction
     VDcopyC(c_grid.VectorLength(), c_u, 0.0);

     i=0;
     do {
     	  MGM(c_grid, c_data, c_ctrl);
	  i++;
	}
     while (( i < f_ctrl->zykluswahl )&&( !c_grid.IsCoarseGrid() ));

//				Interpolate and add correction

     grid.IntAdd(u,c_u);

#ifdef NORM
cout << " correction : " << grid.DefectNorm(u, f)/f0 << endl;
//cout << "            : " << grid.DefectNorm(u, f) <<"  " << f0 << endl;
#endif
//#define DEBUG
#ifdef DEBUG
PrintVector(c_grid.VectorLength(),c_u);
PrintVector(grid.VectorLength(),u);
     grid.Defect(f, u, def);
PrintVector(grid.VectorLength(),def);
#endif

//				Post-Smoothing
     for (i=1; i<=f_ctrl->nGamma; i++)
//	 { grid.GaussSeidelBackward(u,f); }
	 { grid.Smooth(u,f,def,f_ctrl->nGlatt,i); }
#ifdef NORM
cout << " postsmooth : " << grid.DefectNorm(u, f)/f0 << endl;
#endif
   }
 }

//---------------------------------------------------------------------------
void MultigridIteration(unsigned int nlevels,
	const Grid &gf,VectorRec *vecs, CtrlRec *ctrl,
	int &iter, double  &norm)
 {
  double f1,f2,fm,ft;
  double *u   = vecs[nlevels-1].u;
  double *f   = vecs[nlevels-1].f;
  double &eps = ctrl[nlevels-1].eps;
     int nn   = gf.VectorLength();

//    VDcopyC(nn, u, 1.0);		//  Set u=1 on fine grid
//     for (int i=0; i<nn; i++) u[i] = i;
//     cout << "Mult" << endl;
//     gf.Mult(f,u);
//     PrintVector(gf.VectorLength(),f);

 VDcopyC(nn, u, 0.0);		//  Set u=0 on fine grid
//  u[nn/2] = 1.0;

//?? gf.SetBoundaryZero(f);
 f2 = f1 = gf.DefectNorm(u, f);
 fm = f1 * eps;

 iter = 0;
 do
  {
    MGM(gf,&vecs[nlevels-1],&ctrl[nlevels-1]);
    iter++;
    ft = f2;
    f2 = gf.DefectNorm(u, f);
#ifdef ITER_NORM
    cout << endl << "rel. error: " << f2/f1
         << "  (" << f2/ft << ")" << endl;
#endif
  }
 while( (iter<ctrl[nlevels-1].iterat) && (f2>=fm) );

 norm = f2/f1;
// if (norm==0.0)  norm = eps;
 if (abs(norm) < 1e-50)  norm = eps;

 return;
 }

//---------------------------------------------------------------------------
void MultigridIterationNoNorm(unsigned int nlevels,
	const Grid &gf,VectorRec *vecs, CtrlRec *ctrl,
	int &iter)
 {
  double *u   = vecs[nlevels-1].u;
//  double *f   = vecs[nlevels-1].f;
     int nn   = gf.VectorLength();

 VDcopyC(nn, u, 0.0);		//  Set u=0 on fine grid
// gf.SetBoundaryZero(f);

 iter = 0;
 do
  {
    MGM(gf,&vecs[nlevels-1],&ctrl[nlevels-1]);
    ++iter;
  }
 while( iter<ctrl[nlevels-1].iterat );

 return;
 }
//---------------------------------------------------------------------------
void GrobSteuer(CtrlRec &Grob, const CtrlRec &Fein)
 {
   Grob = Fein;
   if (Grob.AllgV)
     {
       Grob.nGamma *= 2;  Grob.vGamma *= 2;
     }
 }
