#ifndef FILE_MULTIGRID_CC
#define FILE_MULTIGRID_CC
#include <cstdlib>
#include <iostream>
using namespace std;
#include "vector.hpp"
#include "grid.hpp"
#include "multigrid.hpp"

#define export
//---------------------------------------------------------------------------
export template<class GRID>
MultiGrid<GRID> :: MultiGrid(const CtrlRec& simple, const unsigned int levels,
		       const GRID& grid_coarse)
 : nlevels(levels), highestlevel(nlevels),
   u_extern(false), f_extern(false),
   grids(0),
//    grids(new GRID[nlevels]),
   ctrl(new CtrlRec[nlevels]), vecs(new VectorRec[nlevels])
{
//   u_extern = f_extern = false;
//   nlevels = levels;
//   highestlevel = nlevels;

  grids = new GRID[levels];
//   ctrl  = new CtrlRec[nlevels];
//   vecs  = new VectorRec[nlevels];

  grids[0] = grid_coarse;

  for (unsigned int i = 1; i < levels; ++i)
  {
    grids[i-1].Refine(grids[i]);
  }
  ctrl[levels-1] = simple;
  for (int i=levels-2; i>=0; --i)	// N O T   unsigend int
     { GrobSteuer(ctrl[i],ctrl[i+1]); }

//			Allocate memory for all arrays of mg
//			u and f on fine grid are external arrays (->true) !!
  ini_level(grids[nlevels-1], vecs[levels-1], ctrl[levels-1], true, true);
  		// Achtung auf Feingrid !!
  for (int i=levels-2; i>=0; --i)	// N O T   unsigend int
    {
      ini_level(grids[i], vecs[i], ctrl[i]);
    }

  cout << endl
       << levels << " Levels ==>  "
       << grids[levels-1].GetIntX() << " intervals in x-dir. ==>  "
       << grids[levels-1].VectorLength() << " unknowns" << endl;
}
//---------------------------------------------------------------------------
export template<class GRID>
MultiGrid<GRID> :: MultiGrid(const CtrlRec& simple, const unsigned int levels,
		       const GRID& grid_coarse, const unsigned int alloclevels)
 : nlevels(min(levels,alloclevels)), highestlevel(alloclevels),
   u_extern(false), f_extern(false),
   grids(new GRID[highestlevel]),
   ctrl(new CtrlRec[highestlevel]), vecs(new VectorRec[highestlevel])
{
//   u_extern     = f_extern = false;
//
//   highestlevel = alloclevels;
//   SetCalcLevel(levels);
//
//   grids = new GRID[highestlevel];
//   ctrl  = new CtrlRec[highestlevel];
//   vecs  = new VectorRec[highestlevel];

  grids[0] = grid_coarse;

  for (unsigned int i = 1; i < highestlevel; ++i)
  {
    grids[i-1].Refine(grids[i]);
  }
  ctrl[highestlevel-1] = simple;
  for (int i=highestlevel-2; i>=0; --i)
     { GrobSteuer(ctrl[i],ctrl[i+1]); }

// 			Allocate memory for all arrays of mg
//			u and f on fine grid are external arrays (->true) !!
//  ini_level(grids[highestlevel-1], vecs[highestlevel-1], ctrl[highestlevel-1], true, true);
// 			Allocate memory for all arrays of mg
//			u and f on fine grid are also internal arrays (->false) !!
  ini_level(grids[highestlevel-1], vecs[highestlevel-1], ctrl[highestlevel-1], false, false);
  		// Achtung auf Feingrid !!
  for (int i=highestlevel-2; i>=0; --i)
    {
      ini_level(grids[i], vecs[i], ctrl[i]);
    }

  cout << endl
       << nlevels << " Levels ==>  "
       << grids[nlevels-1].GetIntX() << " intervals in x-dir. ==>  "
       << grids[nlevels-1].VectorLength() << " unknowns" << endl;
}

//---------------------------------------------------------------------------

// export template<class GRID>
// MultiGrid<GRID> :: MultiGrid(const MultiGrid<GRID> &orig);
// : nlevels(orig.nlevels), highestlevel(orig.highestlevel),
//   u_extern(orig.u_extern), f_extern(orig.f_extern),
//   grids(new GRID[highestlevel]),
//   ctrl(new CtrlRec[highestlevel]), vecs(new VectorRec[highestlevel])
// {}
//---------------------------------------------------------------------------
export template<class GRID>
MultiGrid<GRID> :: ~MultiGrid()
{
//   cout <<  highestlevel<< "   " << nlevels << endl;
 for (unsigned int i=0; i<highestlevel-1; i++)
  {
    delete [] vecs[i].u;
    delete [] vecs[i].f;
    if (vecs[i].def != 0)  delete [] vecs[i].def;
  }
 if (!u_extern) delete [] vecs[highestlevel-1].u;
 if (!f_extern) delete [] vecs[highestlevel-1].f;
 if (vecs[highestlevel-1].def != 0)  delete [] vecs[highestlevel-1].def;

 delete [] vecs;		// Als Klasse! ==>  Destruktor fr pointers
 delete [] ctrl;
 delete [] grids;
}
//---------------------------------------------------------------------------
template<class GRID>
void MultiGrid<GRID> :: SetCalcLevel(const unsigned int level)
{
  if (NumAllocatedLevels() < level)
   {
     cout << " alloclevels < level   in  MultiGrid<GRID> :: SetCalcLevel" << endl;
     exit(-1);
   }
  nlevels = level;
}

//---------------------------------------------------------------------------
export template<class GRID>
void MultiGrid<GRID> :: AddDiagTerm(const double *cc)

{
  GRID      &fine = grids[nlevels-1];
  int  i;

  if ( cc != 0)
   {				//	Adapt for Helmholtz
     if (nlevels>1)
      {
        fine.InitOperator(false, cc );
//DB		cout << "FINE" << endl;PrintVector(fine.VectorLength(),cc);
	fine.Inject(cc, vecs[nlevels-2].u );
//DB		cout << "COARSE" << endl;PrintVector(grids[nlevels-2].VectorLength(),vecs[nlevels-2].u);
	for (i=nlevels-2; i>=1; i--)
	 {
           grids[i].InitOperator(false, vecs[i].u );
	   grids[i].Inject(vecs[i].u, vecs[i-1].u );
//	grids[i].Restrict(vecs[i].u, vecs[i-1].u );
	  }
         grids[0].InitOperator(true, vecs[0].u );
       }
      else
       {
         fine.InitOperator(true, cc );
       }
   }
  else			// Unsauberer HACK
   {
    for (i=nlevels-2; i>=1; i--)
      {
//pp      grids[i].ProjectOperator(false, grids[i+1]);
      }
//pp     grids[0].ProjectOperator(true, grids[1]);
   }
  return;
}
//---------------------------------------------------------------------------

template<class GRID>
void MultiGrid<GRID> ::
	Solve(double u[], double f[], int &iter, double &norm,
			const bool bnorm)

{
  GRID      &fine = grids[nlevels-1];
  VectorRec & fv = vecs[nlevels-1];

//		Copy input data onto MG-data  (improve it!)
  if (u_extern)
  	fv.u = u;
  else
  	VDcopy(fine.VectorLength(), fv.u, 1, u, 1 );
  if (f_extern)
  	fv.f = f;
  else
  	VDcopy(fine.VectorLength(), fv.f, 1, f, 1 );

  if (bnorm)
  	MultigridIteration(nlevels,fine,vecs,ctrl,iter,norm);
  else
   {
  	MultigridIterationNoNorm(nlevels,fine,vecs,ctrl,iter);
	norm = 0.0;
   }

//		Copy MG solution to return vector  (improve it!)
  if (!u_extern)
  	VDcopy(fine.VectorLength(), u, 1, fv.u, 1 );

}
//---------------------------------------------------------------------------
export template<class GRID>
void MultiGrid<GRID> ::
	ini_level(const GRID &grid_l, VectorRec &vecs_l, const CtrlRec &ctrl_l,
		 double *u_in, double *f_in)
 {
   bool bjac,badi;

   if (u_in==0)
    { grid_l.NewGridVector(vecs_l.u); }
   else
    { vecs_l.u = u_in; }

   if (f_in==0)
    { grid_l.NewGridVector(vecs_l.f); }
   else
    { vecs_l.f = f_in; }

/*   Already provided for ResDef without temp. storage of defect */
   bjac = ( ctrl_l.vGlatt == omega_Jacobi || ctrl_l.nGlatt == omega_Jacobi );
   badi = ( ctrl_l.vGlatt == ADI_forw     || ctrl_l.nGlatt == ADI_forw ||
            ctrl_l.vGlatt == ADI_backw    || ctrl_l.nGlatt == ADI_backw );
//bjac=true;
   if (bjac)
   {
     grid_l.NewGridVector(vecs_l.def);

     if (vecs_l.def==0)
      { cout << endl << "Not enough memory for def" << endl; }
   }
   else
    {
     vecs_l.def = 0;
     cout << "Array   def   not needed." << endl;
    }
//				Aux. array for ADI smoothers
//    if (badi && vecs_l.def==0)
//     {
//      int nfx=grid_l.GetIntX(), nfy=grid_l.GetIntY();
// //      data.def  = new double [ (nfx>nfy) ? nfx+1 : nfy+1 ];
// //		data alignement !!
//     }

   return;
 }

//---------------------------------------------------------------------------
export template<class GRID>
void MultiGrid<GRID> ::
	ini_level(const GRID &grid_l, VectorRec &vecs_l, const CtrlRec &ctrl_l,
		  const bool bu, const bool bf)
 {
   bool bjac,badi;

   u_extern = bu;  f_extern = bf;

   if (!u_extern)
    { grid_l.NewGridVector(vecs_l.u); }
   else
    { vecs_l.u = 0; }

   if (!f_extern)
    { grid_l.NewGridVector(vecs_l.f); }
   else
    { vecs_l.f = 0; }

/*   Already provided for ResDef without temp. storage of defect */
   bjac = ( ctrl_l.vGlatt == omega_Jacobi || ctrl_l.nGlatt == omega_Jacobi );
   badi = ( ctrl_l.vGlatt == ADI_forw     || ctrl_l.nGlatt == ADI_forw ||
            ctrl_l.vGlatt == ADI_backw    || ctrl_l.nGlatt == ADI_backw );
//bjac=true;
   if (bjac)
   {
     grid_l.NewGridVector(vecs_l.def);

     if (vecs_l.def==0)
      { cout << endl << "Not enough memory for def" << endl; }
   }
   else
    {
     vecs_l.def = 0;
     cout << "Array   def   not needed." << endl;
    }
//				Aux. array for ADI smoothers
//    if (badi && vecs_l.def==0)
//     {
//      int nfx=grid_l.GetIntX(), nfy=grid_l.GetIntY();
// //     data.def  = new double [ (nfx>nfy) ? nfx+1 : nfy+1 ];
// //		data alignement !!
//     }

   return;
 }
#endif	// MULTIGRID_CC
