#include <cstdlib>
#include <iostream>
#include <string>
#include <cmath>
#include <cstring>
using namespace std;
//#include <ippm.h>

#include "vector.hpp"

//-------------------------------------------------------------------
//		x := y

void VDcopy(const unsigned long n, double *x, const int ix,
			     const double *y, const int iy)
{
  unsigned long i;

  if (ix==1 && iy==1)
    {
      memcpy(x,y,sizeof(double)*n);
    }
  else
    {
     for (i=1; i<=n; *x=*y, ++i, x+=ix, y+=iy) ;
    }
}

//----------------------------------------------------------------------
//		x := const.

void VDcopyC(const unsigned long n, double *x, const double c)
 {
   unsigned long i;

   for (i=0; i<n; x[i]=c, ++i) ;
 }

//----------------------------------------------------------------------
//		x := x / d

void VDdiv(const unsigned long n, double *x, const double *d)
 {
   unsigned long i;

   for (i=0; i<n; x[i] /= d[i], ++i) ;
 }

//----------------------------------------------------------------------
//		y := x / d

void VDdiv(const unsigned long n, double *y, const double *x, const double *d)
 {
   unsigned long i;

   for (i=0; i<n; y[i] = x[i]/d[i], ++i) ;
 }

//----------------------------------------------------------------------
//		x := d * x

void VDscale(const unsigned long n, double *x, const double *d)
 {
   unsigned long i;

   for (i=0; i<n; x[i] *= d[i], ++i) ;
 }

//----------------------------------------------------------------------
//		x := const. * x

void VDscaleC(const unsigned long n, double *x, const double c)
 {
   unsigned long i;

   for (i=0; i<n; x[i] *= c, ++i) ;
 }

//----------------------------------------------------------------------
//		x := x+y

void VDadd(const unsigned long n, double *x, const double *y)
 {
   unsigned long i;

   for (i=0; i<n; x[i] += y[i], ++i) ;
 }

//----------------------------------------------------------------------
//		x := x+y  with strides  ix, iy

void VDadd(const unsigned long n, double *x, const int ix,
			    const double *y, const int iy)
 {
   unsigned long k;
   unsigned long kx,ky;

   for (k=0, kx=0, ky=0; k<=n-1; x[kx] += y[ky], ++k, kx+=ix, ky+=iy) ;
 }

//----------------------------------------------------------------------
//		x := alpha*x + y

void VDaXpY(const unsigned long n, double* x,
		const double alpha, const double* y)
 {
   unsigned long i;
   for (i=0; i<n; x[i] = alpha*x[i] + y[i], ++i) ;
 }

//----------------------------------------------------------------------
//		x := x + alpha*y

void VDXpaY(const unsigned long n, double* x,
		const double alpha, const double* y)
 {
   unsigned long i;
   for (i=0; i<n; x[i] += alpha*y[i], ++i) ;
 }

//----------------------------------------------------------------------
//		z := z + alpha*x + beta*y							*/

void VDaXbYZ(const unsigned long n, const double* x, const double alpha,
	const double* y, const double beta, double* z)
{
 unsigned long i;

 for (i=0; i<n; z[i] += alpha*x[i] + beta*y[i], ++i) ;
}
//----------------------------------------------------------------------
//		L_2 norm of x

double VDnorm(const unsigned long n, double *x)
{
   double        t;
   unsigned long i;

   t = 0.0;
   for (i=0; i<n; ++i)
    {
      t += x[i] * x[i];
    }
   return (sqrt(t));
}
//----------------------------------------------------------------------
double MaxNormRelativeDiff(const unsigned int n, const double* x, const double* y)
{
 double       s,t,tn;
 unsigned int i;

  s = 0.0;
  for (i=0; i<n; i++)
  {
    t  = fabs(x[i]-y[i]);
    tn = 0.5 * ( fabs(x[i]) + fabs(y[i]) );
    if (tn > 1e-20)
       if (t>s*tn)  s = t/tn;
  }

 return s;
}
//---------------------------------------------------------------------
double MaxAbsNorm(const unsigned int n, const double* x)
{
 double       s,t;
 unsigned int i;

  s = 0.0;
  for (i=0; i<n; ++i)
  {
    t  = fabs(x[i]);
    if (t>s)  s = t;
  }

 return s;
}

//----------------------------------------------------------------------
//		inner product of x and y

double VDscalar(const unsigned long n, const double *x, const double *y)
{
   double        t;
   unsigned long i;

//   t = cblas_ddot( n, x, 1, y, 1 ); return t;		//  BLAS in INTEL's mkl-libraries
//   ippmDotProduct_vv_64f( x, y, &t, n );   return t;	//  Intel's IPP-libraries for Itanium2

   t = 0.0;
//#pragma loop count (33*33*33)
//#pragma vector aligned
//#pragma unroll(10)
   for (i=0; i<n; ++i)
    {
      t += x[i] * y[i];
    }
   return (t);
}

//----------------------------------------------------------------------
//		erg := mat * vec

void MatTimesVec(const unsigned long n, const double* mat,
		 const double* vec, double* erg)
{
 unsigned long i,j,ij;

 ij = 0;
 for ( i=0; i<n; ++i)
  {
   erg[i] = 0.0;
   for ( j=0; j<n; ++j)
    {
      erg[i]  +=  mat[ij] * vec[j];
      ij++;
    }
  }

}
//----------------------------------------------------------------------
//		mat := mat^{-1}

void MatrixInversion(const unsigned long n, double* mat)
{

#ifdef LAPACK
 char up = 'U';
 unsigned long i,j;
 int  nn = n, info;

 cout << endl << "Inversion of coarse matrix by LAPACK" << endl;

//			Factor matrix
 dpotrf_( &up, &nn, mat, &nn, &info );
 if (info) cout << endl << "info1 = " << info << "  in InvertMatrix" << endl;

//			Factors --> inverted matrix
 dpotri_( &up, &nn, mat, &nn, &info );
 if (info) cout << endl << "info2 = " << info << "  in InvertMatrix" << endl;

//			Get lower  triangle of inverted matrix
 for (j=0; j<n; j++ )
   for ( i=0; i<j; i++ )
   	mat[i*nn+j] = mat[j*nn+i];

#else
 cout << endl << "No LAPACK used - no exact inversion of Matrix possible." << endl;
 cout << "We will use a diagonal matrix as approximation of inverse." << endl;
 unsigned long i,j,ij;

 ij = 0;
 for ( i=0; i<n; ++i)
  {
   for ( j=0; j<i; ++j)
    {
      mat[ij] = 0.0;
      ij++;
    }
   mat[ij] = 1.0/mat[ij];
   ++ij;
   for ( j=i+1; j<n; ++j)
    {
      mat[ij] = 0.0;
      ++ij;
    }
  }
#endif

}

//----------------------------------------------------------------------

bool CheckMatrixSymmetry(const unsigned long n, const double* mat)
{
 const double EPS_SYMM = 0.5e-5;
 unsigned long int i,j;
 bool bb=true, bj=false;

 cout << endl << "We check the matrix symmetry" << endl;
 for (i=0; i<n; i++)
  {
   for (j=0; j<i; j++)
    {
      bj = (fabs(mat[i*n+j] - mat[j*n+i]) > EPS_SYMM*fabs(mat[i*n+j] + mat[j*n+i]));
      bb = bb && bj;
      if (bj)
        {
         cout << " non-symmetric at (" << i << "," << j << ") : ";
         cout << mat[i*n+j] << "  " << mat[j*n+i] << endl;
        }
    }
  }
 return bb;
}

//----------------------------------------------------------------------

void PrintMatrix(const unsigned long n, const double* mat)
{
 unsigned long int i,j;

 for (i=0; i<n; i++)
  {
   for (j=0; j<n; j++)
    {
      cout << " " << mat[i*n+j];
    }
   cout << endl;
  }
}

//----------------------------------------------------------------------

void PrintVector(const unsigned long n, const double* vec)
{
 const unsigned int NCOL=5;
 unsigned long int i,ii,k;

 ii = 0;
 for (k=0; k<n/NCOL; k++)
   {
     cout << "  " << ii << " : ";
     for (i=0; i<NCOL; i++)
        cout << " " << vec[ii++];
     cout << endl;
   }
 if (ii<n)
   {
    cout << "  " << ii << " : ";
    for (i=ii; i<n; i++)
      {
	cout << " " << vec[i];
      }
     cout << endl;
   }
 cout << endl;
}
//----------------------------------------------------------------------
void SolveTridiag(const unsigned int n, double diag[],
                  double offdiag[], double uf[], const int nrhs)
{
#ifdef LAPACK
//  const int ONE = 1;
        int info;

if (nrhs==1)
{
 FactorSymmTriDiag(n,diag,offdiag);
 SubstFactorSymmTriDiag(n,diag,offdiag,uf);
}
else
{
//  dptsv_(&n, &ONE, diag, offdiag, uf, &n, &info );
  dptsv_(&n, &nrhs, diag, offdiag, uf, &n, &info );
  if (info!=0)
   {
     cout << "Error (" << info << ") in SolveTridiag : " << endl;
     exit(info);
   }
}
#else
 cout << "No SolveTridiag implemented" << endl;
 exit(-1);
#endif
}
//----------------------------------------------------------------------
//	IN  :  A = L + D + L^T
//	OUT :  A = L * D^{-1} * L^T

void FactorSymmTriDiag(const int n, double d[], double l[])
{
 double tmp;
 int   i;

 d[0] = 1.0/d[0];		// inverse of diagonal in factorization

//  for (i=1; i<n; i++)
//   {
//     l[i-1] = l[i-1]*d[i-1];
//     d[i]   = 1.0/(d[i]-l[i-1]*l[i-1]/d[i-1]);
//   }

//	Change in code reduces CPU time from 1.53 sec downto 0.63 sec
//		(cds 9, #define TEST_LaplaceNonEqui2D)

 for (i=1; i<n; ++i)
 {
   tmp = l[i-1];
   l[i-1] = tmp*d[i-1];
   d[i]   = 1.0/(d[i]-l[i-1]*tmp);
 }

 return;
}

//----------------------------------------------------------------------

void SubstFactorSymmTriDiag(const int n, const double d[], const double l[],
				double uf[])
{
 double tmp;
 int    i;

// //   uf[0] = uf[0]
//  for (i=1; i<n; i++)
// 	uf[i] = uf[i] - l[i-1]* uf[i-1];
//
//  for (i=0; i<n; i++)
// 	uf[i] = uf[i]*d[i];

 tmp = uf[0];  uf[0] = tmp*d[0];
 for (i=1; i<n; ++i)
  {
    tmp   = uf[i] - l[i-1]* tmp;
    uf[i] = tmp*d[i];
  }

//   uf[n-1] = uf[n-1];
 for (i=n-2; i>=0; --i)
    uf[i] = uf[i] - l[i]*uf[i+1];

 return;
}

//----------------------------------------------------------------------

void SubstFactorSymmTriDiagT(const int n, const double d[], const double l[],
				double uf[], const int jstride)
{
 double tmp;
 int    i,ik;

 ik = 0;
 tmp = uf[ik];  uf[ik] = tmp*d[0];
 for (i=1; i<n; ++i)
  {
    ik += jstride;
    tmp   = uf[ik] - l[i-1]* tmp;
    uf[ik] = tmp*d[i];
  }

//   uf[n-1] = uf[n-1];
 tmp = uf[ik];
 for (i=n-2; i>=0; --i)
  {
    ik -= jstride;
    uf[ik] = uf[ik] - l[i]*tmp;
    tmp = uf[ik];		// use of tmp accelerates code (20%)
  }

 return;
}

//----------------------------------------------------------------------

void SubstFactorSymmTriDiag_MultT
       (const int n, const double d[], const double l[], double uf[],
	const int nrhs)
{
 int i,k,ik;

//   uf[0] = uf[0]
 for (i=1, ik=nrhs; i<n; ++i)
     for (k=0; k<nrhs; ++k, ++ik)
	uf[ik] = uf[ik] - l[i-1]* uf[ik-nrhs];

//  ==> uf[n-1] = uf[n-1]
 for (ik=(n-1)*nrhs; ik<n*nrhs;  ++ik)
     uf[ik] = uf[ik]*d[n-1];

 for (i=n-2; i>=0; --i)
   {
    ik = i*nrhs;
    for (k=0; k<=nrhs-1; ++k, ++ik)
	uf[ik] = uf[ik]*d[i] - l[i]*uf[ik+nrhs];

   }
 return;
}

/*
//----------------------------------------------------------------------
//	tmp-array resulted in no code acceleration
//
void SubstFactorSymmTriDiag_MultT
       (const int n, const double d[], const double l[], double uf[],
	const int nrhs)
{
 int i,k,ik;
 double  tmp[8193];

//   uf[0] = uf[0]
 ik = 0;
 for (k=0; k<nrhs; k++)  tmp[k] = uf[k];
 for (i=1; i<n; i++)
   {
     ik += nrhs;
     for (k=0; k<nrhs; k++)
       {
// 	uf[ik] = uf[ik] - l[i-1]* uf[ik-nrhs];
 	tmp[k]   = uf[ik+k] - l[i-1]* tmp[k];
	uf[ik+k] = tmp[k]*d[i];
       }
   }

//  ==> uf[n-1] = uf[n-1]

 for (k=0; k<nrhs; k++)  tmp[k] = uf[ik+k];

 for (i=n-2; i>=0; i--)
   {
    ik -= nrhs;
    for (k=0; k<=nrhs-1; k++)
      {
	uf[ik+k] = uf[ik+k] - l[i]*tmp[k];
	tmp[k]   = uf[ik+k];
      }
   }
 return;
}
//----------------------------------------------------------------------
*/
