#ifndef FILE_VECTOR
#define FILE_VECTOR

//	Use LAPACK
// #define LAPACK

#ifdef LAPACK
extern "C"
{
// extern void dpotrf_(const char*, const unsigned long int*, double*, const unsigned long int*, int*);
// extern void dpotri_(const char*, const unsigned long int*, double*, const unsigned long int*, int*);
 extern void dpotrf_(const char*, const int*, double*, const int*, int*);
 extern void dpotri_(const char*, const int*, double*, const int*, int*);
 extern void dptsv_(const unsigned int*, const int*, const double*, const double*,
 		    const double*, const unsigned int*, const int* );

// extern void dpotrf_(char*, int*, double*, int*, int*);
 extern void dsygst_(int*, char*, int*, double*, int*, double*, int*, int*);
 extern void dsytrd_(char*, int*, double*, int*, double*, double*,
					double*, double*, int*, int*);
 extern void dsteqr_(char*, int*, double*, double*, double*,
						int*, double*, int*);
 extern void dsygv_(const int*, const char*, const char*, const int*, double*,
 		const int*, double*, const int*, double*, double*, int*, int*);

 extern double cblas_ddot(const int N, const double *X, const int incX,
 			const double *Y, const int incY);
}
#endif

/**	Vector copy with strides ix and iy
		x := y					*/
void VDcopy(const unsigned long n, double *x, const int ix,
			     const double *y, const int iy);

/**	Copy a constant to a vector
		x := const.				*/
void VDcopyC(const unsigned long n, double *x, const double c);

/**	Scale a vector with inverse elements of a vector
		x := x / d				*/
void VDdiv(const unsigned long n, double *x, const double *d);

/**	Scale a vector with inverse elements of a vector
		y := x / d				*/
void VDdiv(const unsigned long n, double *y, const double *x, const double *d);

/**	Scale a vector with a vector
		x := d * x				*/
void VDscale(const unsigned long n, double *x, const double *d);

/**	Scale a vector with a constant
		x := const. * x				*/
void VDscaleC(const unsigned long n, double *x, const double c);

/**	Add vectors
		x := x + y				*/
void VDadd(const unsigned long n, double *x, const double *y);

/**	Add vectors with strides ix, iy
		x := x + y				*/
void VDadd(const unsigned long n, double *x, const int ix,
                            const double *y, const int iy );

/**		x := alpha*x + y
							*/
void VDaXpY(const unsigned long n, double* x,
		const double alpha, const double* y);

/**		x := x + alpha*y
							*/
void VDXpaY(const unsigned long n, double* x,
		const double alpha, const double* y);

/**		z := z + alpha*x + beta*y
							*/
void VDaXbYZ(const unsigned long n, const double* x, const double alpha,
	const double* y, const double beta, double* z);

/**		L_2 norm of vector
				 			*/
double VDnorm(const unsigned long n, double *x);

/**		abs. Max. norm of the relative difference of two vectors
				 			*/
double MaxNormRelativeDiff(const unsigned int n, const double* x, const double* y);

/**		abs. Max. norm of vector
				 			*/
double MaxAbsNorm(const unsigned int n, const double* x);


/**		inner product of vectors x and y
							*/
double VDscalar(const unsigned long n, const double *x, const double *y);

/**	Matrix-vector product with  a dense square (n x n) matrix
		erg := matrix * vec			*/

void MatTimesVec(const unsigned long n, const double* matrix,
		 const double* vec, double* erg);

/**	Inversion of a dense square (n x n) matrix
		matrix := matrix^{-1}			*/

void MatrixInversion(const unsigned long n, double* matrix);

/**	Checks symmetry of a dense square (n x n) matrix.
							*/
bool CheckMatrixSymmetry(const unsigned long n, const double* mat);

/***	Print a dense square (n x n) matrix.
*/
void PrintMatrix(const unsigned long n, const double* mat);

/**	Print a vector.
*/
void PrintVector(const unsigned long n, const double* vec);

/**		Solves the tridiagonal system  (L+D+L^T) u = f

		Right hand side as well as solution is stored in uf.
		The optional number of right hand sides (nrhs) is
		set to 1 by default.
*/
void SolveTridiag(const unsigned int n, double diag[],
                  double offdiag[], double uf[], const int nrhs=1);


/**		Factors the tridiagonal symm. system matrix.
		diagonal is d[0..n-1], off-diagonal is  l[0..n-2]

		A = L + D + L^T results in  A = (I+L) * D^{-1} * (I+L^T)
*/
void FactorSymmTriDiag(const int n, double d[], double l[]);

/**		Solves [(I+L) * D^{-1} * (I+L^T)] * u = f.

		Vectors f and u are both stored in vector uf.
*/
void SubstFactorSymmTriDiag(const int n, const double d[], const double l[],
				double uf[]);

/**		Solves [(I+L) * D^{-1} * (I+L^T)] * u = f .

		Vectors f and u are both stored in vector uf.
		There is an additional stride for the solution/rhs vector uf.
*/
void SubstFactorSymmTriDiagT(const int n, const double d[], const double l[],
				double uf[], const int jstride);

/**		Solves [(I+L) * D^{-1} * (I+L^T)] * u = f .

		Vectors f and u are both stored in vector uf.
		There are (nrhs) solutions/rhs vectors uf.
		The entries of the (nrhs) vectors belonging to one index are
		stores contignously.
*/
void SubstFactorSymmTriDiag_MultT
       (const int n, const double d[], const double l[], double uf[],
	const int nrhs);

#endif	// FILE_VECTOR
